# Copyright 2019-2020 kubeflow.org
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import json
import re
import pathlib

from typing import List, Optional, Set

from kfp_tekton.compiler._k8s_helper import sanitize_k8s_name
from kfp_tekton.compiler._op_to_template import _get_base_step, _add_mount_path, _prepend_steps, _update_volumes
from os import environ as env

BIG_DATA_MIDPATH = "artifacts/$ORIG_PR_NAME"
BIG_DATA_PATH_FORMAT = "/".join(["$(workspaces.$TASK_NAME.path)", BIG_DATA_MIDPATH, "$TASKRUN_NAME", "$TASK_PARAM_NAME"])
ARTIFACT_OUTPUTLIST_ANNOTATION_KEY = 'artifact_outputs'
OUTPUT_RESULT_PATH_SUFFIX = '-datap'


def fix_big_data_passing(workflow: dict, loops_pipeline: dict, loop_name_prefix: str) -> dict:
    """
    fix_big_data_passing converts a workflow where some artifact data is passed
    as parameters and converts it to a workflow where this data is passed as
    artifacts.

    Args:
        workflow: The workflow to fix
    Returns:
        The fixed workflow

    Motivation:

    DSL compiler only supports passing Tekton parameters.
    Due to the convoluted nature of the DSL compiler, the artifact consumption
    and passing has been implemented on top of that using parameter passing.
    The artifact data is passed as parameters and the consumer template creates
    an artifact/file out of that data.
    Due to the limitations of Kubernetes and Tekton this scheme cannot pass data
    larger than few kilobytes preventing any serious use of artifacts.

    This function rewrites the compiled workflow so that the data consumed as
    artifact is passed as artifact.
    It also prunes the unused parameter outputs. This is important since if a
    big piece of data is ever returned through a file that is also output as
    parameter, the execution will fail.
    This makes is possible to pass large amounts of data.

    Implementation:

    1. Index the pipelines to understand how data is being passed and which inputs/outputs
       are connected to each other.
    2. Search for direct data consumers in container/resource templates and some
       pipeline task attributes (e.g. conditions and loops) to find out which inputs
       are directly consumed as parameters/artifacts.
    3. Propagate the consumption information upstream to all inputs/outputs all
       the way up to the data producers.
    4. Convert the inputs, outputs based on how they're consumed downstream.
    5. Use workspaces instead of result and params for big data passing.
    6. Added workspaces to tasks, pipelines, pipelineruns, if the parmas is big data.
    7. A PVC named with pipelinerun name will be created if big data is passed, as workspaces need to use it.
       User need to define proper volume or enable dynamic volume provisioning refer to the link of:
       https://kubernetes.io/docs/concepts/storage/dynamic-provisioning
    """

    workflow = copy.deepcopy(workflow)

    tasks = workflow["spec"]["pipelineSpec"].get(
        'tasks', []) + workflow["spec"]["pipelineSpec"].get('finally', [])

    resource_templates = []
    for task in tasks:
        resource_params = [
            param.get('name') for param in task.get("taskSpec", {}).get('params', [])
            if param.get('name') == 'action'
            or param.get('name') == 'success-condition'
        ]
        if 'action' in resource_params and 'success-condition' in resource_params:
            resource_templates.append(task)

    # TODO: use split on condition
    resource_template_names = set(task["name"] for task in resource_templates)

    container_templates = [
        task for task in tasks if task["name"] not in resource_template_names
    ]

    pipeline_template = workflow["spec"]["pipelineSpec"]

    pipelinerun_template = load_annotations(workflow)

    # 1. Index the pipelines to understand how data is being passed and which
    #  inputs/outputs are connected to each other.
    template_input_to_parent_pipeline_inputs = {
    }  # (task_template_name, task_input_name) -> Set[(pipeline_template_name, pipeline_input_name)]
    template_input_to_parent_task_outputs = {
    }  # (task_template_name, task_input_name) -> Set[(upstream_template_name, upstream_output_name)]
    template_input_to_parent_constant_arguments = {
    }  # (task_template_name, task_input_name) -> Set[argument_value] # Unused
    pipeline_output_to_parent_template_outputs = {
    }  # (pipeline_template_name, output_name) -> Set[(upstream_template_name, upstream_output_name)]

    # pipeline has no name when embedded in pipelineRun
    pipelinerun_name = workflow.get('metadata', {}).get('name')

    # Indexing task arguments
    pipeline_tasks = pipeline_template.get(
        "tasks", []) + pipeline_template.get('finally', [])

    for task in pipeline_tasks:
        task_template_name = task['name']
        parameter_arguments = task['params']
        for parameter_argument in parameter_arguments:
            task_input_name = parameter_argument['name']
            argument_value = parameter_argument['value']

            argument_placeholder_parts = deconstruct_tekton_single_placeholder(
                argument_value)
            if not argument_placeholder_parts:  # Argument is considered to be constant string
                template_input_to_parent_constant_arguments.setdefault(
                    (task_template_name, task_input_name),
                    set()).add(argument_value)
            else:
                placeholder_type = argument_placeholder_parts[0]
                if placeholder_type not in ('params', 'outputs', 'tasks',
                                            'steps', 'workflow', 'pod',
                                            'item'):
                    # Do not fail on Jinja or other double-curly-brace templates
                    continue
                if placeholder_type == 'params':
                    pipeline_input_name = argument_placeholder_parts[1]
                    template_input_to_parent_pipeline_inputs.setdefault(
                        (task_template_name, task_input_name), set()).add(
                            (pipelinerun_name, pipeline_input_name))
                elif placeholder_type == 'tasks':
                    upstream_task_name = argument_placeholder_parts[1]
                    assert argument_placeholder_parts[2] == 'results'
                    upstream_output_name = argument_placeholder_parts[3]
                    # upstream_template_name = upstream_task_name
                    template_input_to_parent_task_outputs.setdefault(
                        (task_template_name, task_input_name), set()).add(
                            (upstream_task_name, upstream_output_name))
                elif placeholder_type in ('item', 'workflow', 'pod'):
                    # workflow.parameters.* placeholders are not supported,
                    # but the DSL compiler does not produce those.
                    template_input_to_parent_constant_arguments.setdefault(
                        (task_template_name, task_input_name),
                        set()).add(argument_value)
                else:
                    raise AssertionError
            pipeline_input_name = extract_tekton_input_parameter_name(
                argument_value)
            if pipeline_input_name:
                template_input_to_parent_pipeline_inputs.setdefault(
                    (task_template_name, task_input_name), set()).add(
                        (pipelinerun_name, pipeline_input_name))
            else:
                template_input_to_parent_constant_arguments.setdefault(
                    (task_template_name, task_input_name),
                    set()).add(argument_value)
    # Finshed indexing the pipelines

    # 2. Search for direct data consumers in container/resource templates and some pipeline task attributes
    #  (e.g. conditions and loops) to find out which inputs are directly consumed as parameters/artifacts.
    inputs_directly_consumed_as_parameters = set()
    inputs_directly_consumed_as_artifacts = set()
    outputs_directly_consumed_as_parameters = set()

    # Searching for artifact input consumers in container template inputs
    for template in container_templates:
        template_name = template['name']
        for input_artifact in template.get("taskSpec", {}).get('artifacts', {}):
            raw_data = input_artifact['raw'][
                'data']  # The structure must exist
            # The raw data must be a single input parameter reference. Otherwise (e.g. it's a string
            #  or a string with multiple inputs) we should not do the conversion to artifact passing.
            input_name = extract_tekton_input_parameter_name(raw_data)
            if input_name:
                inputs_directly_consumed_as_artifacts.add(
                    (template_name, input_name))
                del input_artifact[
                    'raw']  # Deleting the "default value based" data passing hack
                # so that it's replaced by the "argument based" way of data passing.
                input_artifact[
                    'name'] = input_name  # The input artifact name should be the same
                # as the original input parameter name

    # Searching for parameter input consumers in container and resource templates
    for template in container_templates + resource_templates:
        template_name = template['name']
        placeholders = extract_all_tekton_placeholders(
            template.get('taskSpec', {}))
        for placeholder in placeholders:
            parts = placeholder.split('.')
            placeholder_type = parts[0]
            if placeholder_type not in ('inputs', 'outputs', 'tasks', 'steps',
                                        'workflow', 'pod', 'item', 'results'):
                # Do not fail on Jinja or other double-curly-brace templates
                continue

            if placeholder_type == 'workflow' or placeholder_type == 'pod':
                pass
            elif placeholder_type == 'inputs':
                if parts[1] == 'params':
                    input_name = parts[2]
                    inputs_directly_consumed_as_parameters.add(
                        (template_name, input_name))
                elif parts[1] == 'artifacts':
                    raise RuntimeError(
                        'Found unexpected Tekton input artifact placeholder in container template: {}'
                        .format(placeholder))
                else:
                    raise RuntimeError(
                        'Found unexpected Tekton input placeholder in container template: {}'
                        .format(placeholder))
            elif placeholder_type == 'results':
                input_name = parts[1]
                outputs_directly_consumed_as_parameters.add(
                    (template_name, input_name))
            else:
                raise RuntimeError(
                    'Found unexpected Tekton placeholder in container template: {}'
                    .format(placeholder))

    # Finished indexing data consumers

    # 3. Propagate the consumption information upstream to all inputs/outputs all the way up to the data producers.
    inputs_consumed_as_parameters = set()
    inputs_consumed_as_artifacts = set()

    outputs_consumed_as_parameters = set()
    outputs_consumed_as_artifacts = set()

    def mark_upstream_ios_of_input(template_input, marked_inputs,
                                   marked_outputs):
        # Stopping if the input has already been visited to save time and handle recursive calls
        if template_input in marked_inputs:
            return
        marked_inputs.add(template_input)

        upstream_inputs = template_input_to_parent_pipeline_inputs.get(
            template_input, [])
        for upstream_input in upstream_inputs:
            mark_upstream_ios_of_input(upstream_input, marked_inputs,
                                       marked_outputs)

        upstream_outputs = template_input_to_parent_task_outputs.get(
            template_input, [])
        for upstream_output in upstream_outputs:
            mark_upstream_ios_of_output(upstream_output, marked_inputs,
                                        marked_outputs)

    def mark_upstream_ios_of_output(template_output, marked_inputs,
                                    marked_outputs):
        # Stopping if the output has already been visited to save time and handle recursive calls
        if template_output in marked_outputs:
            return
        marked_outputs.add(template_output)

        upstream_outputs = pipeline_output_to_parent_template_outputs.get(
            template_output, [])
        for upstream_output in upstream_outputs:
            mark_upstream_ios_of_output(upstream_output, marked_inputs,
                                        marked_outputs)

    for input in inputs_directly_consumed_as_parameters:
        mark_upstream_ios_of_input(input, inputs_consumed_as_parameters,
                                   outputs_consumed_as_parameters)
    for input in inputs_directly_consumed_as_artifacts:
        mark_upstream_ios_of_input(input, inputs_consumed_as_artifacts,
                                   outputs_consumed_as_artifacts)
    for output in outputs_directly_consumed_as_parameters:
        mark_upstream_ios_of_output(output, inputs_consumed_as_parameters,
                                    outputs_consumed_as_parameters)

    # 4. Convert the inputs, outputs and arguments based on how they're consumed downstream.
    # Add workspaces to pipeline and pipeline task_ref if big data passing
    pipeline_workspaces = set()
    pipelinerun_workspaces = set()
    output_tasks_consumed_as_artifacts = {
        output[0]
        for output in outputs_consumed_as_artifacts
    }
    # task_workspaces = set()

    # Converting pipeline inputs
    pipeline, pipeline_workspaces = big_data_passing_pipeline(
        pipelinerun_name, pipeline_template, inputs_consumed_as_artifacts,
        output_tasks_consumed_as_artifacts)

    # Add workspaces to pipelinerun if big data passing
    if pipeline_workspaces:
        pipeline, pipelinerun_workspaces = big_data_passing_pipelinerun(
            pipelinerun_name, pipelinerun_template, pipeline_workspaces)

    # Use workspaces to tasks if big data passing instead of 'results', 'copy-inputs'
    for task_template in container_templates:
        task_template = big_data_passing_tasks(pipelinerun_name,
                                               task_template,
                                               pipelinerun_template,
                                               inputs_consumed_as_artifacts,
                                               outputs_consumed_as_artifacts,
                                               loops_pipeline,
                                               loop_name_prefix)

    # Remove input parameters unless they're used downstream.
    # This also removes unused container template inputs if any.
    for template in container_templates + [pipeline_template]:
        spec = template.get('taskSpec', {}) or template.get('pipelineSpec', {})
        spec['params'] = [
            input_parameter for input_parameter in spec.get('params', []) if (
                template.get(
                    'name'
                ),  # TODO: pipeline has no name, use pipelineRun name?
                input_parameter['name']) in inputs_consumed_as_parameters
                or input_parameter['name'].endswith("-trname")
                or input_parameter['name'].endswith(OUTPUT_RESULT_PATH_SUFFIX)
        ]

    # Remove output parameters unless they're used downstream
    for template in container_templates + [pipeline_template]:
        spec = template.get('taskSpec', {}) or template.get('pipelineSpec', {})
        spec['results'] = [
            output_parameter for output_parameter in spec.get('results', [])
        ]
        spec['results'] = sorted(spec['results'], key=lambda k: k['name'])
        # tekton results doesn't support underscore
        renamed_results_in_pipeline_task = set()
        for task_result in spec['results']:
            task_result_old_name = task_result.get('name')
            task_result_new_name = sanitize_k8s_name(task_result_old_name, allow_capital=True)
            if task_result_new_name != task_result_old_name:
                task_result['name'] = task_result_new_name
                renamed_results_in_pipeline_task.add(
                    (task_result_old_name, task_result_new_name))
        for renamed_result in renamed_results_in_pipeline_task:
            # Change results.downloaded_resultOutput to results.downloaded-resultoutput
            template['taskSpec'] = replace_big_data_placeholder(
                spec, 'results.%s' % renamed_result[0],
                'results.%s' % renamed_result[1])

    # Remove pipeline task parameters unless they're used downstream
    for task in pipeline_tasks:
        # Don't process condition and custom task parameters
        is_condition = 'condition-' in task['name']
        is_custom_task = task.get('taskRef') or task.get('taskSpec', {}).get('apiVersion')
        if not is_condition and not is_custom_task:
            task['params'] = [
                parameter_argument
                for parameter_argument in task.get('params', [])
                if (task['name'], parameter_argument['name']
                    ) in inputs_consumed_as_parameters and
                (task['name'],
                parameter_argument['name']) not in inputs_consumed_as_artifacts
                or task['name'] in resource_template_names
                or parameter_argument['name'].endswith("-trname")
                or parameter_argument['name'].endswith(OUTPUT_RESULT_PATH_SUFFIX)
            ]

        # tekton results doesn't support underscore
        for argument in task['params']:
            argument_value = argument.get('value')
            argument_placeholder_parts = deconstruct_tekton_single_placeholder(
                argument_value)
            if len(argument_placeholder_parts) == 4 \
                    and argument_placeholder_parts[0] == 'tasks':
                argument['value'] = '$(tasks.%s.%s.%s)' % (
                    argument_placeholder_parts[1],
                    argument_placeholder_parts[2],
                    sanitize_k8s_name(argument_placeholder_parts[3], allow_capital=True))

    workflow = jsonify_annotations(workflow)
    # Need to confirm:
    # I didn't find the use cases to support workflow parameter consumed as artifacts downstream in tekton.
    # Whether this case need to be supporting?
    clean_up_empty_workflow_structures(workflow)
    return workflow


def extract_all_tekton_placeholders(template: dict) -> Set[str]:
    template_str = json.dumps(template)
    placeholders = set(re.findall('\\$\\(([-._a-zA-Z0-9]+)\\)', template_str))
    return placeholders


def extract_tekton_input_parameter_name(s: str) -> Optional[str]:
    match = re.fullmatch('\\$\\(inputs.params.([-_a-zA-Z0-9]+)\\)', s)
    if not match:
        return None
    (input_name, ) = match.groups()
    return input_name


def deconstruct_tekton_single_placeholder(s: str) -> List[str]:
    if not re.fullmatch('\\$\\([-._a-zA-Z0-9]+\\)', s):
        return []
    return s.lstrip('$(').rstrip(')').split('.')


def replace_big_data_placeholder(template: dict, old_str: str,
                                 new_str: str) -> dict:
    template_str = json.dumps(template)
    template_str = template_str.replace(old_str, new_str)
    template = json.loads(template_str)
    return template


def big_data_passing_pipeline(name: str, template: dict, inputs_tasks: set(),
                              outputs_tasks: set):
    pipeline_workspaces = set()
    pipeline_name = name
    pipeline_spec = template
    tasks = pipeline_spec.get('tasks', []) + pipeline_spec.get('finally', [])
    for task in tasks:
        parameter_arguments = task.get('params', [])
        for parameter_argument in parameter_arguments:
            input_name = parameter_argument['name']
            if (task.get('name'), input_name) in inputs_tasks:
                pipeline_workspaces.add(pipeline_name)
                # Add workspaces instead of params, for tasks of big data inputs
                if not task.setdefault('workspaces', []):
                    task['workspaces'].append({
                        "name": task.get('name'),
                        "workspace": pipeline_name
                    })
                # runAfter add to the task, which was depends on the result of the task.
                argument_value = parameter_argument['value']
                argument_placeholder_parts = deconstruct_tekton_single_placeholder(
                    argument_value)
                if argument_placeholder_parts[
                        0] == 'tasks' and argument_placeholder_parts[
                            2] == 'results':
                    dependency_task = argument_placeholder_parts[1]
                    task.setdefault('runAfter', [])
                    task['runAfter'].append(dependency_task)
        if task.get('name') in outputs_tasks:
            # Add workspaces for tasks of big data outputs
            if not task.setdefault('workspaces', []):
                task['workspaces'].append({
                    "name": task.get('name'),
                    "workspace": pipeline_name
                })
        artifact_output_list = task.get('taskSpec', {}).get('metadata', {}).get('annotations', {}).get(
            ARTIFACT_OUTPUTLIST_ANNOTATION_KEY, '')
        if artifact_output_list:
            tmp_list = set()
            for output in json.loads(artifact_output_list):
                tmp_list.add(sanitize_k8s_name(output, allow_capital=True))
            for task_output in task.get('taskSpec', {}).get('results', []):
                if task_output.get('name') in tmp_list:
                    if not task.setdefault('workspaces', []):
                        task['workspaces'].append({
                          "name": task.get('name'),
                            "workspace": pipeline_name
                        })
                        pipeline_workspaces.add(pipeline_name)
                        break

    if pipeline_name in pipeline_workspaces:
        # Add workspaces to pipeline
        if not pipeline_spec.setdefault('workspaces', []):
            pipeline_spec['workspaces'].append({"name": pipeline_name})
    return template, pipeline_workspaces


def big_data_passing_pipelinerun(name: str, pr: dict, pw: set):
    prw = set()
    pipelinerun_name = name
    if pipelinerun_name in pw:
        pr.get('spec', {}).setdefault('workspaces', [])
        # Change persistentVolumeClaim to volumeClaimTemplate which need Tekton version > 0.12
        # User could modify default size of storage in the yaml file mannually if necessary.
        DEFAULT_ACCESSMODES = env.get('DEFAULT_ACCESSMODES', 'ReadWriteMany')
        DEFAULT_STORAGE_SIZE = env.get('DEFAULT_STORAGE_SIZE', '2Gi')
        DEFAULT_STORAGE_CLASS = env.get('DEFAULT_STORAGE_CLASS', 'kfp-csi-s3')
        pr['spec']['workspaces'].append({
            "name": pipelinerun_name,
            "volumeClaimTemplate": {
                "spec": {
                    "storageClassName": DEFAULT_STORAGE_CLASS,
                    "accessModes": [DEFAULT_ACCESSMODES],
                    "resources": {
                        "requests": {
                            "storage": DEFAULT_STORAGE_SIZE
                        }
                    }
                }
            }
        })
        prw.add(pipelinerun_name)
    return pr, prw


def big_data_passing_tasks(prname: str, task: dict, pipelinerun_template: dict,
                            inputs_tasks: set, outputs_tasks: set, loops_pipeline: dict,
                            loop_name_prefix: str) -> dict:
    task_name = task.get('name')
    task_spec = task.get('taskSpec', {})
    # Data passing for the task outputs
    appended_taskrun_name = False
    appended_taskrun_path_step = None
    artifact_output_list = task_spec.get('metadata', {}).get('annotations', {}).get(ARTIFACT_OUTPUTLIST_ANNOTATION_KEY, '')
    if artifact_output_list:
        temp_list = json.loads(artifact_output_list)
        artifact_output_list = []
        for output in temp_list:
            artifact_output_list.append(sanitize_k8s_name(output, allow_capital=True))
    for task_output in task.get('taskSpec', {}).get('results', []):
        if (task_name, task_output.get('name')) in outputs_tasks or \
            (artifact_output_list and task_output.get('name') in artifact_output_list):
            if not task.get('taskSpec', {}).setdefault('workspaces', []):
                task.get('taskSpec', {})['workspaces'].append({"name": task_name})
            # Replace the args for the outputs in the task_spec
            # $(results.task_output.get('name').path)  -->
            # $(workspaces.task_name.path)/task_name-task_output.get('name')
            placeholder = '$(results.%s.path)' % (sanitize_k8s_name(
                task_output.get('name'), allow_capital=True))
            workspaces_parameter = '$(workspaces.%s.path)/%s/%s/%s' % (
                task_name, BIG_DATA_MIDPATH, "$(context.taskRun.name)", task_output.get('name'))
            if env.get('OUTPUT_BIG_DATA_PATH', 'false').lower() == 'true':
            # For child nodes to know the taskrun output path, it has to pass to results via /tekton/results emptydir
                if not appended_taskrun_path_step:
                    appended_taskrun_path_step = _get_base_step('output-taskrun-path')
                if len(appended_taskrun_path_step['command']) <= 2:
                    appended_taskrun_path_step['command'].append('')
                appended_taskrun_path_step['command'][-1] += 'echo -n "%s/%s/%s" > $(results.%s%s.path)\n' % \
                                                    (BIG_DATA_MIDPATH, "$(context.taskRun.name)", task_output.get('name'),
                                                    task_output.get('name'), OUTPUT_RESULT_PATH_SUFFIX)
                task['taskSpec']['results'].append({"name": "%s%s" % (task_output.get('name'), OUTPUT_RESULT_PATH_SUFFIX),
                                                    "type": "string"})
            else:
                # For child nodes to know the taskrun name, it has to pass to results via /tekton/results emptydir
                if not appended_taskrun_name:
                    copy_taskrun_name_step = _get_base_step('output-taskrun-name')
                    copy_taskrun_name_step['command'].append('echo -n "$(context.taskRun.name)" > "$(results.taskrun-name.path)"')
                    task['taskSpec']['results'].append({"name": "taskrun-name", "type": "string"})
                    task['taskSpec']['steps'].append(copy_taskrun_name_step)
                    _append_original_pr_name_env(task)
                    appended_taskrun_name = True
            task['taskSpec'] = replace_big_data_placeholder(
                task.get("taskSpec", {}), placeholder, workspaces_parameter)
            artifact_items = pipelinerun_template['metadata']['annotations'].get('tekton.dev/artifact_items', '{}')
            if artifact_items != '{}':
                artifact_items[task['name']] = replace_big_data_placeholder(
                    artifact_items[task['name']], placeholder, workspaces_parameter)
                pipelinerun_template['metadata']['annotations']['tekton.dev/artifact_items'] = \
                    artifact_items
    if appended_taskrun_path_step:
        task['taskSpec']['steps'].append(appended_taskrun_path_step)
        _append_original_pr_name_env(task)

    task_spec = task.get('taskSpec', {})
    task_params = task_spec.get('params', [])
    task_artifacts = task_spec.get('artifacts', [])

    # Data passing for task inputs
    for task_param in task_params:
        if (task_name, task_param.get('name')) in inputs_tasks:
            if not task_spec.setdefault('workspaces', []):
                task_spec['workspaces'].append({"name": task_name})
            # Replace the args for the inputs in the task_spec
            # /tmp/inputs/text/data ---->
            # $(workspaces.task_name.path)/task_param.get('name')
            placeholder = '/tmp/inputs/text/data'
            for task_artifact in task_artifacts:
                if task_artifact.get('name') == task_param.get('name'):
                    placeholder = task_artifact.get('path')
            task_param_task_name = ""
            task_param_param_name = ""
            for o_task in outputs_tasks:
                if '-'.join(o_task) == task_param.get('name'):
                    task_param_task_name = o_task[0]
                    task_param_param_name = o_task[1]
                    break
            # If the param name is constructed with task_name-param_name,
            # use the current task_name as the path prefix

            def append_taskrun_params(task_name_append: str, task_path_name: str):
                taskrun_param_name = task_name_append + "-trname"
                inserted_taskrun_param = False
                param_name = {"name": taskrun_param_name}
                param_nested_name = task_name_append + '-taskrun-name'
                param_content = {"name": taskrun_param_name, "value": "$(tasks.%s.results.taskrun-name)" % task_name_append}
                param_nested_content = {'name': task_name_append + '-taskrun-name',
                                        'value': '$(tasks.%s.results.taskrun-name)' % task_name_append}
                if env.get('OUTPUT_BIG_DATA_PATH', 'false').lower() == 'true':
                    param_path_name = '-'.join([task_name_append, task_path])
                    param_name = {"name": param_path_name}
                    param_nested_name = param_path_name
                    param_content = {"name": param_path_name,
                                     "value": "$(tasks.%s.results.%s)" % (task_name_append, task_path_name)}
                    param_nested_content = param_content
                else:
                    for param in task['taskSpec'].get('params', []):
                        if param.get('name', "") == taskrun_param_name:
                            inserted_taskrun_param = True
                            break
                if (not inserted_taskrun_param) or env.get('OUTPUT_BIG_DATA_PATH', 'false').lower() == 'true':
                    task['taskSpec']['params'].append(param_name)
                    task['params'].append(param_content)
                    parent_task_queue = [task['name']]
                    while parent_task_queue:
                        current_task = parent_task_queue.pop(0)
                        for loop_name, loop_spec in loops_pipeline.items():
                            if current_task in loop_spec.get('task_list', []):
                                parent_task_queue.append(loop_name.replace(loop_name_prefix, ""))
                                loop_param_names = [loop_param['name'] for loop_param in loops_pipeline[loop_name]['spec']['params']]
                                if param_nested_name in loop_param_names:
                                    continue
                                loops_pipeline[loop_name]['spec']['params'].append(param_nested_content)

            if task_param_task_name:
                workspaces_parameter = '$(workspaces.%s.path)/%s/$(params.%s-trname)/%s' % (
                    task_name, BIG_DATA_MIDPATH, task_param_task_name, task_param_param_name)
                task_path = sanitize_k8s_name(task_param_param_name, allow_capital=True) + OUTPUT_RESULT_PATH_SUFFIX
                if env.get('OUTPUT_BIG_DATA_PATH', 'false').lower() == 'true':
                    workspaces_parameter = '$(workspaces.%s.path)/$(params.%s)' % (task_name, '-'.join([task_param_task_name, task_path]))
                if task_param_task_name != task_name:
                    append_taskrun_params(task_param_task_name, task_path)  # need to get taskrun name from parent path
            else:
                workspaces_parameter = '$(workspaces.%s.path)/%s/%s/%s' % (
                    task_name, BIG_DATA_MIDPATH, "$(context.taskRun.name)", task_param.get('name'))
            _append_original_pr_name_env(task)
            task['taskSpec'] = replace_big_data_placeholder(
                task_spec, placeholder, workspaces_parameter)
            task_spec = task.get('taskSpec', {})
    # Handle the case of input artifact without dependent the output of other tasks
    for task_artifact in task_artifacts:
        if (task_name, task_artifact.get('name')) not in inputs_tasks:
            # add input artifact processes
            task = input_artifacts_tasks(task, task_artifact)

        if (prname, task_artifact.get('name')) in inputs_tasks:
            # add input artifact processes for pipeline parameter
            if not task_artifact.setdefault('raw', {}):
                for i in range(len(pipelinerun_template['spec']['params'])):
                    param_name = pipelinerun_template['spec']['params'][i]['name']
                    param_value = pipelinerun_template['spec']['params'][i]['value']
                    if (task_artifact.get('name') == param_name):
                        task_artifact['raw']['data'] = param_value
                        task = input_artifacts_tasks_pr_params(task, task_artifact)

    # If a task produces a result and artifact, add a step to copy artifact to results.
    artifact_items = pipelinerun_template['metadata']['annotations'].get('tekton.dev/artifact_items', '{}')
    add_copy_results_artifacts_step = False
    if task.get("taskSpec", {}):
        if task_spec.get('results', []):
            copy_results_artifact_step = _get_base_step('copy-results-artifacts')
            copy_results_artifact_step['onError'] = 'continue'  # supported by v0.27+ of tekton.
            script = "set -exo pipefail\nTOTAL_SIZE=0\n"
            injected_script = False
            for result in task_spec['results']:
                if task['name'] in artifact_items:
                    artifact_i = artifact_items[task['name']]
                    for index, artifact_tuple in enumerate(artifact_i):
                        artifact_name, artifact = artifact_tuple
                        src = artifact
                        dst = '$(results.%s.path)' % sanitize_k8s_name(result['name'], allow_capital=True)
                        if artifact_name == result['name'] and src != dst:
                            if not injected_script:
                                add_copy_results_artifacts_step = True
                                src_arg = '"$1"'
                                dst_arg = '"$2"'
                                total_size_command = 'ARTIFACT_SIZE=`wc -c %s${SUFFIX} | awk \'{print $1}\'`\n' % src_arg + \
                                                    'TOTAL_SIZE=$( expr $TOTAL_SIZE + $ARTIFACT_SIZE)\n'
                                copy_command = '    cp ' + src_arg + ' ' + dst_arg + '\n'
                                if env.get('OUTPUT_PREVIEW', 'false').lower() == 'true':
                                    preview_size = env.get('OUTPUT_PREVIEW_SIZE', '100')
                                    total_size_command = 'TOTAL_SIZE=$( expr $TOTAL_SIZE + %s)\n' % preview_size
                                    copy_command = '    dd if=' + src_arg + ' of=' + \
                                                dst_arg + ' bs=' + preview_size + ' count=1\n'
                                script += (
                                        'copy_artifact() {\n'
                                        'if [ -d ' + src_arg + ' ]; then\n' +
                                        '  tar -czvf ' + src_arg + '.tar.gz ' + src_arg + '\n' +
                                        '  SUFFIX=".tar.gz"\n' +
                                        'fi\n' +
                                        total_size_command +
                                        'touch ' + dst_arg + '\n' +  # create an empty file by default.
                                        'if [[ $TOTAL_SIZE -lt 3072 ]]; then\n' +
                                        '  if [ -d ' + src_arg + ' ]; then\n' +
                                        '    tar -tzf ' + src_arg + '.tar.gz > ' + dst_arg + '\n' +
                                        '  elif ! awk "/[^[:print:]]/{f=1} END{exit !f}" %s; then\n' % src_arg +
                                        copy_command +
                                        '  fi\n' +
                                        'fi\n' +
                                        '}\n'
                                )
                                injected_script = True
                            script += 'copy_artifact %s %s\n' % (src, dst)
            copy_results_artifact_step['command'].append(script)
            _append_original_pr_name_env_to_step(copy_results_artifact_step)
            if add_copy_results_artifacts_step:
                task['taskSpec']['steps'].append(copy_results_artifact_step)

    # Remove artifacts parameter from params
    task.get("taskSpec", {})['params'] = [
        param for param in task_spec.get('params', [])
        if (task_name, param.get('name')) not in inputs_tasks or
            param.get('name').endswith("-trname") or
            param.get('name').endswith(OUTPUT_RESULT_PATH_SUFFIX)
    ]

    # Remove artifacts from task_spec
    if 'artifacts' in task_spec:
        del task['taskSpec']['artifacts']

    return task


def input_artifacts_tasks_pr_params(template: dict, artifact: dict) -> dict:
    copy_inputs_step = _get_base_step('copy-inputs')
    task_name = template.get('name')
    task_spec = template.get('taskSpec', {})
    task_params = task_spec.get('params', [])
    script = "set -exo pipefail\n"
    for task_param in task_params:
        # For pipeline parameter input artifacts, it will never come from another task because pipeline
        # params are global parameters. Thus, task_name will always be the executing task name.
        workspaces_parameter = '$(workspaces.%s.path)/%s/%s/%s' % (
            task_name, BIG_DATA_MIDPATH, "$(context.taskRun.name)", task_param.get('name'))
        if 'raw' in artifact:
            script += 'mkdir -p %s\n' % pathlib.Path(workspaces_parameter).parent
            script += 'echo -n "%s" > %s\n' % (
                artifact['raw']['data'], workspaces_parameter)
        _append_original_pr_name_env(template)

    copy_inputs_step['command'].append(script)
    template['taskSpec']['steps'] = _prepend_steps(
        [copy_inputs_step], template['taskSpec']['steps'])

    return template


def input_artifacts_tasks(template: dict, artifact: dict) -> dict:
    # The input artifacts in KFP is not pulling from s3, it will always be passed as a raw input.
    # Visit https://github.com/kubeflow/pipelines/issues/336 for more details on the implementation.
    volume_mount_step_template = []
    volume_template = []
    mounted_param_paths = []
    copy_inputs_step = _get_base_step('copy-inputs')
    has_copy_input = False
    copy_input_index = 0
    if 'raw' in artifact:
        for step_index in range(len(template['taskSpec']['steps'])):
            if template['taskSpec']['steps'][step_index].get('name') == 'copy-inputs':
                copy_input_index = step_index
                has_copy_input = True
        if has_copy_input:
            template['taskSpec']['steps'][copy_input_index]['command'][-1] = \
            template['taskSpec']['steps'][copy_input_index]['command'][-1] + \
            'echo -n "%s" > %s\n' % (artifact['raw']['data'], artifact['path'])
        else:
            copy_inputs_step['command'].append('set -exo pipefail\necho -n "%s" > %s\n' % (
                artifact['raw']['data'], artifact['path']))
    mount_path = artifact['path'].rsplit("/", 1)[0]
    if mount_path not in mounted_param_paths:
        _add_mount_path(artifact['name'], artifact['path'], mount_path,
                        volume_mount_step_template, volume_template,
                        mounted_param_paths)
    if not has_copy_input:
        template['taskSpec']['steps'] = _prepend_steps(
            [copy_inputs_step], template['taskSpec']['steps'])
    _update_volumes(template['taskSpec'], volume_mount_step_template, volume_template)
    return template


def clean_up_empty_workflow_structures(workflow: list):
    template_spec = workflow['spec']
    if not template_spec.setdefault('params', []):
        del template_spec['params']
    if not template_spec.setdefault('artifacts', []):
        del template_spec['artifacts']
    if not template_spec.setdefault('results', []):
        del template_spec['results']
    if not template_spec:  # ?
        del workflow['spec']
    # if template['kind'] == 'Pipeline':
    for task in workflow['spec']['pipelineSpec']['tasks'] + workflow['spec'][
            'pipelineSpec'].get('finally', []):
        if not task.setdefault('params', []):
            del task['params']
        if not task.setdefault('artifacts', []):
            del task['artifacts']
    if not workflow['spec']['pipelineSpec'].setdefault('finally', []):
        del workflow['spec']['pipelineSpec']['finally']


def load_annotations(template: dict):
    if env.get('DISABLE_ARTIFACT_TRACKING', 'false').lower() != 'true':
        artifact_items = json.loads(
            str(template['metadata']['annotations'].get('tekton.dev/artifact_items', '{}')))
        template['metadata']['annotations']['tekton.dev/artifact_items'] = \
            artifact_items
    return template


def jsonify_annotations(template: dict):
    if env.get('DISABLE_ARTIFACT_TRACKING', 'false').lower() != 'true':
        template['metadata']['annotations']['tekton.dev/artifact_items'] = \
            json.dumps(template['metadata']['annotations'].get('tekton.dev/artifact_items', '{}'))
    return template


def _append_original_pr_name_env_to_step(step):
    step.setdefault('env', [])
    has_original_pr_name = False
    for task_env in step['env']:
        if task_env['name'] == "ORIG_PR_NAME":
            has_original_pr_name = True
    if not has_original_pr_name:
        step['env'].append({"name": "ORIG_PR_NAME",
                            "valueFrom":
                                {"fieldRef":
                                    {"fieldPath": "metadata.labels['custom.tekton.dev/originalPipelineRun']"}}})


def _append_original_pr_name_env(task_template):
    for step in task_template['taskSpec']['steps']:
        if step['name'] == 'main':
            _append_original_pr_name_env_to_step(step)


def fix_big_data_passing_using_volume(workflow, pipeline_conf):
    if workflow['spec'].get('workspaces') and pipeline_conf.data_passing_method._volume is not None:
        volume_dict = pipeline_conf.data_passing_method._volume.to_dict()
        volume_dict_pvc = volume_dict.get('persistent_volume_claim')
        if volume_dict_pvc:
            for workspace in workflow['spec']['workspaces']:
                if workspace['name'] == workflow['metadata']['name']:
                    workspace.pop('volumeClaimTemplate')
                    temp_vc = {}
                    for key, value in volume_dict_pvc.items():
                        if value != None:
                            if key == 'claim_name':
                                temp_vc['claimName'] = value
                            if key == 'read_only':
                                temp_vc['readOnly'] = value
                    workspace['persistentVolumeClaim'] = temp_vc
                    if pipeline_conf.data_passing_method._path_prefix != None:
                        workspace['subPath'] = pipeline_conf.data_passing_method._path_prefix
                    break
    return workflow



# Copyright 2021 kubeflow.org
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import List, Iterable, Union, Optional, TypeVar, Text, Tuple, Dict

from kfp.dsl import _pipeline_param, _for_loop, _pipeline
from kfp import dsl
from kfp import components
from kfp.dsl._for_loop import LoopArguments, ItemList
from kfp.dsl._pipeline_param import ConditionOperator, PipelineParam
from kfp_tekton.compiler._k8s_helper import sanitize_k8s_name

BREAK_TASK_IMAGE_NAME = "aipipeline/breaktask:latest"
CEL_EVAL_IMAGE = "aipipeline/cel-eval:latest"
ANY_SEQUENCER_IMAGE = "quay.io/aipipeline/any-sequencer:latest"
DEFAULT_CONDITION_OUTPUT_KEYWORD = "outcome"
TEKTON_CUSTOM_TASK_IMAGES = [CEL_EVAL_IMAGE, BREAK_TASK_IMAGE_NAME]
LOOP_PIPELINE_NAME_LENGTH = 40
LOOP_GROUP_NAME_LENGTH = 16
_Num = TypeVar('_Num', int, float)


def AnySequencer(any: Iterable[Union[dsl.ContainerOp, ConditionOperator]],
                name: str = None, statusPath: str = None,
                skippingPolicy: str = None, errorPolicy: str = None,
                image: str = ANY_SEQUENCER_IMAGE):
    """A containerOp that will proceed when any of the dependent containerOps completed
       successfully

    Args:
        name: The name of the containerOp. It does not have to be unique within a pipeline
                because the pipeline will generate a unique new name in case of conflicts.

        any: List of `Conditional` containerOps that deploy together with the `main`
                containerOp, or the condtion that must meet to continue.

        statusPath: The location to write the output stauts

        skippingPolicy: Determines for the Any Sequencer reacts to
                no-dependency-condition-matching case. Values can be one of `skipOnNoMatch`
                or `errorOnNoMatch`, a status with value "Skipped" will be generated and the
                exit status will still be succeeded on `skipOnNoMatch`.

        errorPolicy: The standard field, either `failOnError` or `continueOnError`. On
                `continueOnError`, a status with value "Failed" will be generated
                but the exit status will still be succeeded. For `Fail_on_error` the
                Any Sequencer should truly fail in the Tekton terms, as it does now.

        image: The image to implement the any sequencer logic. Default to aipipeline/any-sequencer:latest.
    """
    arguments = [
                "--namespace",
                "$(context.pipelineRun.namespace)",
                "--prName",
                "$(context.pipelineRun.name)"
            ]
    tasks_list = []
    condition_list = []
    file_outputs = None
    for cop in any:
        if isinstance(cop, dsl.ContainerOp):
            cop_name = sanitize_k8s_name(cop.name)
            tasks_list.append(cop_name)
        elif isinstance(cop, ConditionOperator):
            condition_list.append(cop)
    if len(tasks_list) > 0:
        task_list_str = "\'" + ",".join(tasks_list) + "\'"
        arguments.extend(["--taskList", task_list_str])
    if statusPath is not None:
        file_outputs = '{outputPath: %s}' % statusPath
        arguments.extend(["--statusPath", file_outputs])
        if skippingPolicy is not None:
            assert skippingPolicy == "skipOnNoMatch" or skippingPolicy == "errorOnNoMatch"
            arguments.extend(["--skippingPolicy", skippingPolicy])
        if errorPolicy is not None:
            assert errorPolicy == "continueOnError" or errorPolicy == "failOnError"
            arguments.extend(["--errorPolicy", errorPolicy])
    conditonArgs = processConditionArgs(condition_list)
    arguments.extend(conditonArgs)

    AnyOp_yaml = '''\
    name: %s
    description: 'Proceed when any of the dependents completed successfully'
    outputs:
    - {name: %s, description: 'The output file to create the status'}
    implementation:
        container:
            image: %s
            command: [any-task]
            args: [%s]
    ''' % (name, statusPath, image, ",".join(arguments))
    AnyOp_template = components.load_component_from_text(AnyOp_yaml)
    AnyOp = AnyOp_template()
    return AnyOp


def processOperand(operand) -> (str, str):
    if isinstance(operand, dsl.PipelineParam):
        return "results_" + sanitize_k8s_name(operand.op_name) + "_" + \
                sanitize_k8s_name(operand.name, allow_capital=True), operand.op_name
    else:
        # Do the same as in _get_super_condition_template to check whehter it's int
        try:
            operand = int(operand)
        except:
            operand = '\'' + str(operand) + '\''
        return operand, None


def processCondition(condition: ConditionOperator) -> str:
    op1, taskName1 = processOperand(condition.operand1)
    op2, taskName2 = processOperand(condition.operand2)
    if taskName1 is not None and taskName2 is not None:
        assert taskName1 == taskName2, "The result for condition must come from the same task."
    assert taskName1 is not None or taskName2 is not None, "Must at least contain one result in one condition for a task."
    conditionStr = f"{op1} {condition.operator} {op2}"
    return conditionStr


def processConditionArgs(conditions: List[ConditionOperator]) -> List[str]:
    conditionArgs = []
    for condition in conditions:
        conditionStr = processCondition(condition)
        conditionArgs.extend(["-c", conditionStr])
    return conditionArgs


def after_any(any: Iterable[Union[dsl.ContainerOp, ConditionOperator]], name: str = None,
              statusPath: str = None, skippingPolicy: str = None, errorPolicy: str = None):
    '''
    The function adds a new AnySequencer and connects the given op to it
    '''
    seq = AnySequencer(any, name, statusPath, skippingPolicy, errorPolicy)

    def _after_components(cop):
        cop.after(seq)
        return cop
    return _after_components


def CEL_ConditionOp(condition_statement):
    '''A containerOp template for CEL and converts it into Tekton custom task
    during Tekton compiliation.

    Args:
        condition_statement: CEL expression statement using string and/or pipeline params.
    '''
    ConditionOp_yaml = '''\
    name: 'condition-cel'
    description: 'Condition Operation using Common Expression Language'
    inputs:
    - {name: condition_statement, type: String, description: 'Condition statement', default: ''}
    outputs:
    - {name: %s, type: String, description: 'Default condition output'}
    implementation:
        container:
            image: %s
            command: ['sh', '-c']
            args: [
            '--apiVersion', 'cel.tekton.dev/v1alpha1',
            '--kind', 'CEL',
            '--name', 'cel_condition',
            '--%s', {inputValue: condition_statement},
            {outputPath: %s},
            ]
    ''' % (DEFAULT_CONDITION_OUTPUT_KEYWORD,
           CEL_EVAL_IMAGE,
           DEFAULT_CONDITION_OUTPUT_KEYWORD,
           DEFAULT_CONDITION_OUTPUT_KEYWORD)
    ConditionOp_template = components.load_component_from_text(ConditionOp_yaml)
    ConditionOp = ConditionOp_template(condition_statement)
    return ConditionOp


def Break():
    '''A BreakOp template for Break Operation using PipelineLoop
    '''
    BreakOp_yaml = '''\
    name: 'pipelineloop-break-operation'
    description: 'Break Operation using PipelineLoop'
    implementation:
        container:
            image: %s
            command: ['sh', '-c']
            args: [
             '--apiVersion', 'custom.tekton.dev/v1alpha1',
             '--kind', 'BreakTask',
             '--name', 'pipelineloop-break-operation',
            ]
    ''' % (BREAK_TASK_IMAGE_NAME)
    BreakOp_template = components.load_component_from_text(BreakOp_yaml)
    BreakOp = BreakOp_template()
    return BreakOp


class TektonLoopArguments(LoopArguments):
  def __init__(self,
               items: Union[str, ItemList, dsl.PipelineParam],
               code: Text,
               name_override: Optional[Text] = None,
               op_name: Optional[Text] = None,
               *args,
               **kwargs):
    if isinstance(items, str):
        # temporary list wrapping for validation to pass
        super().__init__([items], code, name_override, op_name, *args, **kwargs)
        self.items_or_pipeline_param = items
    else:
        super().__init__(items, code, name_override, op_name, *args, **kwargs)

  def to_str_for_task_yaml(self):
    if isinstance(self.items_or_pipeline_param, str):
      return self.items_or_pipeline_param
    else:
      raise ValueError(
          'You should only call this method on loop args which are string literals, '
          'not lists or pipeline param items.')


class TektonLoopIterationNumber(PipelineParam):
    def __init__(self, name: str):
        super().__init__(name=name, param_type='Integer')


class Loop(dsl.ParallelFor):

  ITERATION_NUMBER = 'iteration-number'

  @classmethod
  def sequential(cls,
                 loop_args: _for_loop.ItemList,
                 extra_fields: Optional[dict] = None,
                 valid_extra_field_values: Optional[dict] = None):
    return cls(loop_args=loop_args,
               parallelism=1,
               extra_fields=extra_fields,
               valid_extra_field_values=valid_extra_field_values)

  @classmethod
  def from_string(cls,
                  loop_args: Union[str, _pipeline_param.PipelineParam],
                  separator: Optional[Union[str, _pipeline_param.PipelineParam]] = None,
                  parallelism: Optional[int] = None,
                  extra_fields: Optional[dict] = None,
                  valid_extra_field_values: Optional[dict] = None):
    return cls(loop_args=loop_args,
               separator=separator,
               parallelism=parallelism,
               extra_fields=extra_fields,
               valid_extra_field_values=valid_extra_field_values)

  @classmethod
  def range(cls,
            start: Union[_Num, PipelineParam],
            end: Union[_Num, PipelineParam],
            step: Optional[Union[_Num, PipelineParam]] = None,
            parallelism: Optional[int] = None,
            extra_fields: Optional[dict] = None,
            valid_extra_field_values: Optional[dict] = None):
    return cls(start=start,
               step=step,
               end=end,
               parallelism=parallelism,
               extra_fields=extra_fields,
               valid_extra_field_values=valid_extra_field_values)

  def add_pod_annotation(self, name: str, value: str):
    """Adds a pod's metadata annotation.
    Args:
        name: The name of the annotation.
        value: The value of the annotation.
    """

    self.pod_annotations[name] = value
    return self

  def add_pod_label(self, name: str, value: str):
    """Adds a pod's metadata label.
    Args:
        name: The name of the label.
        value: The value of the label.
    """

    self.pod_labels[name] = value
    return self

  def _next_id(self):
    return str(_pipeline.Pipeline.get_default_pipeline().get_next_group_id())

  def _seek_id(self):
    return str(_pipeline.Pipeline.get_default_pipeline().group_id)

  def __init__(self,
               loop_args: Union[str,
                                _for_loop.ItemList,
                                _pipeline_param.PipelineParam] = None,
               start: Union[_Num, PipelineParam, None] = None,
               end: Union[_Num, PipelineParam, None] = None,
               step: Union[_Num, PipelineParam, None] = None,
               separator: Optional[Union[str, _pipeline_param.PipelineParam]] = None,
               parallelism: Optional[int] = None,
               extra_fields: Optional[dict] = None,
               valid_extra_field_values: Optional[dict] = None):
    self.start = None
    self.end = None
    self.step = None
    self.call_enumerate = False
    self.iteration_number = None
    self.pod_annotations = {}
    self.pod_labels = {}
    self.iterate_param_pass_style = None
    self.item_pass_style = None
    self.config_value_list = {"iterate_param_pass_style": ["inline", "file"],
                              "item_pass_style": ["inline", "file"]}
    if extra_fields:
        if extra_fields.get('iterate_param_pass_style'):
            self.iterate_param_pass_style = extra_fields['iterate_param_pass_style']
        if extra_fields.get('item_pass_style'):
            self.item_pass_style = extra_fields['item_pass_style']

    # Update allowed values in the extra fields
    if valid_extra_field_values:
        for k, v in valid_extra_field_values.items():
            self.config_value_list[k] = v

    if start and end:
        super().__init__(loop_args=["iteration"], parallelism=parallelism)
        self.start = start
        self.end = end
        self.step = step
    else:
        if loop_args is None and (start is None or end is None):
            raise RuntimeError("loop_args or start/end parameters are missing for 'Loop' class")

        if isinstance(loop_args, str):
            # temporary list wrapping for validation to pass
            super().__init__(loop_args=[loop_args], parallelism=parallelism)
            self.loop_args = TektonLoopArguments(
                loop_args,
                code=self._next_id(),
            )
            self.items_is_string = True
        else:
            super().__init__(loop_args=loop_args, parallelism=parallelism)
            self.loop_args = TektonLoopArguments(
                loop_args,
                code=self._next_id(),
            )
            self.items_is_string = False

        self.separator = None
        if separator is not None:
            self.separator = PipelineParam(
                name=LoopArguments._make_name(self._next_id()),
                value=separator
            )

    # param_name is unique on the pipeline level but not cluster level. Therefore, we still need to replace it to a cluster
    # unique uuid during our compiler step.
    param_name = "-".join([_pipeline.Pipeline.get_default_pipeline().name, self.type, str(int(self._seek_id()) + 1)])
    self.last_idx = dsl.PipelineParam(name="last-idx", op_name=param_name)
    self.last_elem = dsl.PipelineParam(name="last-elem", op_name=param_name)

  def __enter__(self) -> Union[Tuple[TektonLoopIterationNumber, LoopArguments], _for_loop.LoopArguments]:
    rev = super().__enter__()
    if self.call_enumerate:
      self.iteration_number = TektonLoopIterationNumber(name=self.ITERATION_NUMBER + '-' + self._next_id())
      return (self.iteration_number, self.loop_args)
    return rev

  def enumerate(self) -> dsl.ParallelFor:
    self.call_enumerate = True
    return self


class AddOnGroup(dsl.OpsGroup):
    """
    Represents an AddOn group containing ops and group of OpsGroups.
    This class is the base class for a customized OpsGroup. Users can
    develop their own OpsGroup. The customized OpsGroup maps to a
    custom task in Tekton.
    """
    TYPE_NAME = 'addon_group'
    TASK_TYPE = 'task'
    RUN_TYPE = 'run'  # means custom task
    DEFAULT_KIND = 'AddOnGroup'
    DEFAULT_APIVERSION = 'custom.tekton.dev/v1alpha1'

    @classmethod
    def create_internal_param(cls, name: str, value: Optional[str] = None, param_type: Optional[str] = None) -> dsl.PipelineParam:
        """
        If a PipelineParam is used for underlying custom task and should not show up in spec.params, use this
        function to create a PipelineParam and add it into the AddOnGroup.params.
        """
        rev = dsl.PipelineParam(name=name, value=value, param_type=param_type)
        rev.addon_param = True
        return rev

    def __init__(self, task_type: str = RUN_TYPE,
                kind: str = DEFAULT_KIND,
                api_version: str = DEFAULT_APIVERSION,
                is_finally: bool = False,
                parallelism: int = None,
                params: Dict[str, Union[dsl.PipelineParam, str, int]] = {},
                annotations: Dict[str, str] = {},
                labels: Dict[str, str] = {}):
        self.task_type = task_type  # not been used yet
        self.kind = kind
        self.api_version = api_version
        self.params = params  # for spec.params
        self.annotations = annotations  # for metadata.annotation
        self.labels = labels  # for metadata.labels
        if is_finally:
            pl = dsl._pipeline.Pipeline.get_default_pipeline()
            if pl.groups[-1].type != 'pipeline':
                raise ValueError(
                    'You can only create a finally OpsGroup under the root OpsGroup of a Pipeline')
        self.is_finally = is_finally  # a regular task or finally task
        super().__init__(self.TYPE_NAME, parallelism=parallelism)

    def post_task_spec(self, task_yaml: dict) -> dict:
        """provide a post-hook api to update the task YAML"""
        return task_yaml

    def post_params(self, params: List):
        """provide a post-hook api to update params"""
        return params



# Copyright 2019-2021 kubeflow.org.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json, copy, re
from kfp_tekton.compiler._k8s_helper import sanitize_k8s_name

from kfp_tekton.tekton import LOOP_GROUP_NAME_LENGTH, AddOnGroup


def _process_argo_vars(workflow):
    # Use regex to replace all the Argo variables to Tekton variables. For variables that are unique to Argo,
    # we raise an Error to alert users about the unsupported variables. Here is the list of Argo variables.
    # https://github.com/argoproj/argo/blob/master/docs/variables.md
    # Since Argo variables can be used in anywhere in the yaml, we need to dump and then parse the whole yaml
    # using regular expression.
    tekton_var_regex_rules = [
        {
        'argo_rule': '{{inputs.parameters.([^ \t\n.:,;{}]+)}}',
        'tekton_rule': '$(inputs.params.\g<1>)'
        },
        {
        'argo_rule': '{{outputs.parameters.([^ \t\n.:,;{}]+).path}}',
        'tekton_rule': '$(results.\g<1>.path)'
        },
        {
        'argo_rule': '{{workflow.uid}}',
        'tekton_rule': '$(context.pipelineRun.uid)'
        },
        {
        'argo_rule': '{{workflow.name}}',
        'tekton_rule': '$(context.pipelineRun.name)'
        },
        {
        'argo_rule': '{{workflow.namespace}}',
        'tekton_rule': '$(context.pipelineRun.namespace)'
        },
        {
        'argo_rule': '{{workflow.parameters.([^ \t\n.:,;{}]+)}}',
        'tekton_rule': '$(params.\g<1>)'
        }
    ]
    for regex_rule in tekton_var_regex_rules:
        workflow = re.sub(regex_rule['argo_rule'], regex_rule['tekton_rule'], workflow)
    return workflow


def _handle_tekton_pipeline_variables(pipeline_run):
    """
    Handle tekton pipeline level variables, such as context.pipelineRun.name
    See more here https://github.com/tektoncd/pipeline/blob/master/docs/variables.md
    If there are tekton pipeline level variables in container, then
    1. Replease the $(context.pipeline/pipelineRun).*) to $(params.var_name)
       for example context.pipelineRun.name -> $(params.pipelineRun-name)
    2. Add $(context.pipeline/pipelineRun).*) to pipeline_run['spec']['pipelineSpec']['tasks']['params']
    3. Add $(context.pipeline/pipelineRun).*) to pipeline_run['spec']['pipelineSpec']['tasks']['taskSpec']['params']
    """

    pipeline_variables = {
        'pipeline-name': '$(context.pipeline.name)',
        'pipelineRun-name': '$(context.pipelineRun.name)',
        'pipelineRun-namespace': '$(context.pipelineRun.namespace)',
        'pipelineRun-uid': '$(context.pipelineRun.uid)'
    }

    add_type = pipeline_run.get('kind') != 'PipelineRun'

    def handle_task(task):
        if task.get('taskRef', {}):
            return
        if 'taskSpec' in task and 'apiVersion' in task['taskSpec']:
            # recur for embedded pipeline resources
            if 'spec' in task['taskSpec'] and 'pipelineSpec' in task['taskSpec']['spec']:
                resource_spec = task['taskSpec']
                _handle_tekton_pipeline_variables(resource_spec)
            return

        for key, val in pipeline_variables.items():
            task_str = json.dumps(task['taskSpec']['steps'])
            task_str = _process_argo_vars(task_str)
            if val in task_str:
                task_str = task_str.replace(val, '$(params.' + key + ')')
                task['taskSpec']['steps'] = json.loads(task_str)
                if task.get('params', ''):
                    if {'name': key, 'value': val} not in task['params']:
                        task['params'].append({'name': key, 'value': val})
                else:
                    task['params'] = [{'name': key, 'value': val}]

                found = False
                for param in task['taskSpec'].get('params', []):
                    if param['name'] == key:
                        found = True
                        break
                if not found:
                    param = {'name': key}
                    if add_type:
                        param['type'] = 'string'
                    task['taskSpec'].setdefault('params', []).append(param)

    task_list = pipeline_run.get('spec', {}).get('pipelineSpec', {}).get('tasks', [])
    finally_task_list = pipeline_run.get('spec', {}).get('pipelineSpec', {}).get('finally', [])
    for task in task_list:
        handle_task(task)
    for task in finally_task_list:
        handle_task(task)

    return pipeline_run


def _handle_tekton_custom_task(custom_task: dict, workflow: dict, recursive_tasks: list, group_names: list):
    """
    Separate custom task's workflow from the main workflow, return a tuple result of custom task cr definitions
    and a new workflow

    Args:
      custom_task: dictionary with custom_task infomation, the format should be as below:
      {
        'kind': '',
        'task_list': [],
        'spec': {},
        'depends': []
      }
      workflow: a workflow without loop pipeline separeted.
      recursive_tasks: List of recursive_tasks information.
      group_names: List of name constructions for creating custom loop crd names.

    Returns:
      A tuple (custom_task_crs, workflow).
      custom_task_crs is a list of custom task cr definitions.
      and workflow is a dict which will not including the tasks in custom task definitions
    """
    custom_task_crs = []
    task_list = []
    tasks = workflow['spec']['pipelineSpec'].get('tasks', [])
    new_tasks = []
    dependencies = []
    # handle dependecies
    for key in custom_task.keys():
        dependencies.extend(custom_task[key]['depends'])
    for task in tasks:
        for dependency in dependencies:
            if task['name'] == dependency['org']:
                task_dependencies = [dependency['runAfter']]
                for dep_task in task.get('runAfter', []):
                    # should exclude the custom task itself for cases like graph
                    dep_task_trim = copy.copy(dep_task)
                    if len(group_names[-1]) <= LOOP_GROUP_NAME_LENGTH:
                        dep_task_trim = sanitize_k8s_name(dep_task, max_length=LOOP_GROUP_NAME_LENGTH, rev_truncate=True)
                    dep_task_with_prefix = '-'.join(group_names[:-1] + [dep_task_trim])
                    if dep_task_with_prefix == dependency['runAfter']:
                        continue
                    if dep_task not in custom_task[dependency['runAfter']]['task_list']:
                        task_dependencies.append(dep_task)
                task['runAfter'] = task_dependencies

    # process recursive tasks to match parameters
    for task in recursive_tasks:
        recursive_graph = custom_task.get(task['taskRef']['name'], {})
        if recursive_graph:
            if recursive_graph['spec']['params']:
                recursive_graph['spec']['params'] = sorted(recursive_graph['spec']['params'], key=lambda k: k['name'])
            for param in recursive_graph['spec']['params']:
                recursive_params = [param['name'] for param in task['params']]
                if param['name'] not in recursive_params:
                    task['params'].append({'name': param['name'], 'value': "$(params.%s)" % param['name']})

    # get custom tasks
    for custom_task_key in custom_task.keys():
        denpendency_list = custom_task[custom_task_key]['spec'].get('runAfter', [])
        task_list.extend(custom_task[custom_task_key]['task_list'])
        # generate custom task cr
        custom_task_cr_tasks = []
        for task in tasks:
            if task['name'] in custom_task[custom_task_key]['task_list']:
                for param in task.get('taskSpec', {}).get('params', []):
                    param['type'] = 'string'
                run_after_task_list = []
                for run_after_task in task.get('runAfter', []):
                    for recursive_task in recursive_tasks:
                        # The subset of the loop group name should be LOOP_GROUP_NAME_LENGTH minus 4 because the
                        # numbers of loop cannot exceed 1000 due to ETCD limitation.
                        if sanitize_k8s_name(recursive_task['name'], max_length=(LOOP_GROUP_NAME_LENGTH - 4), rev_truncate=True) \
                            in run_after_task and '-'.join(group_names[:-1]) not in run_after_task:
                            if len(group_names[-1]) <= LOOP_GROUP_NAME_LENGTH:
                                run_after_task = sanitize_k8s_name(run_after_task, max_length=LOOP_GROUP_NAME_LENGTH, rev_truncate=True)
                            run_after_task = '-'.join(group_names[:-1] + [run_after_task])
                            break
                    if run_after_task not in denpendency_list:
                        # check task name is task list
                        for origin_task_name in task_list:
                            if origin_task_name in run_after_task:
                                run_after_task_list.append(run_after_task)
                                break
                if task.get('runAfter', []):
                    task['runAfter'] = run_after_task_list
                custom_task_cr_tasks.append(task)
        # append recursive tasks
        for task in recursive_tasks:
            if task['name'] in custom_task[custom_task_key]['task_list']:
                custom_task_cr_tasks.append(task)
        # generator custom task cr
        custom_task_cr = {
            "apiVersion": "custom.tekton.dev/v1alpha1",
            "kind": 'custom_task_kind',
            "metadata": {
                "name": custom_task_key
            },
            "spec": {
                "pipelineSpec": {
                    "params": [{
                        "name": parm['name'],
                        'type': 'string'
                    } for parm in sorted(custom_task[custom_task_key]['spec']['params'], key=lambda k: k['name'])],
                    "tasks": custom_task_cr_tasks
                }
            }
        }

        # handle loop special case
        if custom_task[custom_task_key]['kind'] == 'loops':
            # if subvar exist, this is dict loop parameters
            # remove the loop_arg and add subvar args to the cr params
            if custom_task[custom_task_key]['loop_sub_args'] != []:
                refesh_cr_params = []
                for param in custom_task_cr['spec']['pipelineSpec']['params']:
                    if param['name'] != custom_task[custom_task_key]['loop_args']:
                        refesh_cr_params.append(param)
                custom_task_cr['spec']['pipelineSpec']['params'] = refesh_cr_params
                custom_task_cr['spec']['pipelineSpec']['params'].extend([{
                    "name": sub_param,
                    'type': 'string'
                } for sub_param in custom_task[custom_task_key]['loop_sub_args']])

            # add loop special fields
            custom_task_cr['kind'] = 'PipelineLoop'

            def process_inline_cr_field(field_name):
                if custom_task[custom_task_key]['spec'].get(field_name) is not None:
                    custom_task_cr['spec'][field_name] = custom_task[custom_task_key]['spec'][field_name]
                    # remove from pipeline run spec
                    del custom_task[custom_task_key]['spec'][field_name]
            process_fields = ['parallelism', 'iterateParamPassStyle', 'itemPassStyle']
            for process_field in process_fields:
                process_inline_cr_field(process_field)
            if custom_task[custom_task_key].get('iteration_number') is not None:
                # enumerate() is used, TektonLoopIterationNumber is injected, get its name and remove it from params
                custom_task_cr['spec']['iterationNumberParam'] = custom_task[custom_task_key]['iteration_number']
                custom_task[custom_task_key]['spec']['params'] = [custom_task_param for custom_task_param
                                                                  in custom_task[custom_task_key]['spec']['params']
                                                                  if custom_task_param['name'] !=
                                                                      custom_task_cr['spec']['iterationNumberParam']]
            custom_task_cr['spec']['iterateParam'] = custom_task[custom_task_key]['loop_args']
            separator = custom_task[custom_task_key].get('separator')
            start_end_step_keys = ['from', 'to', 'step']
            if separator is not None:
                custom_task_cr['spec']['iterateParamStringSeparator'] = separator
            if custom_task[custom_task_key].get('start') is not None:
                custom_task_cr['spec']['pipelineSpec']['params'] = [value for value
                                                                    in custom_task_cr['spec']['pipelineSpec']['params']
                                                                    if value['name'] not in start_end_step_keys]
                custom_task[custom_task_key]['spec']['params'] = [custom_task_param for custom_task_param
                                                                  in custom_task[custom_task_key]['spec']['params']
                                                                  if custom_task_param['name'] != custom_task_cr['spec']['iterateParam']]

                custom_task_cr['spec']['iterateNumeric'] = custom_task_cr['spec']['iterateParam']
                custom_task_cr['spec'].pop('iterateParam')
            # check whether or not the nested custom task param reference need to be replaced
            custom_task_param_map = {}
            for custom_task_param in custom_task[custom_task_key]['spec']['params']:
                if custom_task_param['name'] != custom_task[custom_task_key]['loop_args'] and '$(tasks.' in custom_task_param['value']:
                    custom_task_param_map.setdefault(custom_task_param['value'], []).append('$(params.%s)' % custom_task_param['name'])
            for key, item in custom_task_param_map.items():
                replacement_item = None
                if len(item) == 1:
                    replacement_item = item[0]
                if len(item) > 1:
                    forbidden_keystrings = ['$(params.%s)' % x for x in start_end_step_keys]
                    for i in item:
                        if i not in forbidden_keystrings:
                            replacement_item = i
                            break
                if replacement_item:
                    custom_task_cr = json.loads(
                        json.dumps(custom_task_cr).replace(key, replacement_item))

            # remove separator from CR params
            if custom_task[custom_task_key].get('separator') is not None:
                separator_param = custom_task[custom_task_key]['separator']
                custom_task_cr['spec']['pipelineSpec']['params'] = [
                    param
                    for param in custom_task_cr['spec']['pipelineSpec']['params']
                    if param['name'] != separator_param
                ]

        if custom_task[custom_task_key]['kind'] == 'addon':
            # remove params that are created for internal use. those params have `addon_param` attribute
            custom_task[custom_task_key]['spec']['params'] = [custom_task_param for custom_task_param
                in custom_task[custom_task_key]['spec']['params']
                if not hasattr(custom_task[custom_task_key]['_data'].params.get(custom_task_param['name'], {}), 'addon_param')]
            # clean up params in spec.taskSpec.spec.pipelineSpec.params, remove those
            # are not used by downstream Ops/OpsGroups. one-level down only
            child_task_params = []
            for task in custom_task_cr['spec']['pipelineSpec']['tasks']:
                for task_param in task.get('params', []):
                    if task_param['value'].startswith('$(') and task_param['value'].endswith(')'):
                        child_task_params.append(task_param['name'])
            custom_task_cr['spec']['pipelineSpec']['params'] = [value for value
                                                    in custom_task_cr['spec']['pipelineSpec']['params']
                                                    if value['name'] in child_task_params]

        # need to process task parameters to replace out of scope results
        # because nested graph cannot refer to task results outside of the sub-pipeline.
        custom_task_cr_task_names = [custom_task_cr_task['name'] for custom_task_cr_task in custom_task_cr['spec']['pipelineSpec']['tasks']]
        for task in custom_task_cr['spec']['pipelineSpec']['tasks']:
            for task_param in task.get('params', []):
                if '$(tasks.' in task_param['value']:
                    param_results = re.findall('\$\(tasks.([^ \t\n.:,;\{\}]+).results.([^ \t\n.:,;\{\}]+)\)', task_param['value'])
                    for param_result in param_results:
                        if param_result[0] not in custom_task_cr_task_names:
                            if param_result[0] == 'None':
                                task['params'] = json.loads(
                                    json.dumps(task['params']).replace(task_param['value'],
                                    '$(params.%s)' % param_result[1]))
                            else:
                                new_param_ref = '%s-%s' % param_result
                                # Nested task reference name must be the same as the one in the pipeline params
                                # Otherwise, use the default param ref since it will be generated
                                # by the custom task functions such as condition and loop
                                for i in custom_task_cr['spec']['pipelineSpec'].get('params', []):
                                    if sanitize_k8s_name(i['name']) == new_param_ref:
                                        new_param_ref = i['name']
                                        break
                                task['params'] = json.loads(
                                    json.dumps(task['params']).replace(task_param['value'],
                                    '$(params.%s)' % new_param_ref))
        custom_task_crs.append(custom_task_cr)
        custom_task[custom_task_key]['spec']['params'] = sorted(custom_task[custom_task_key]['spec']['params'],
                                                                          key=lambda k: k['name'])
        tasks.append(custom_task[custom_task_key]['spec'])

    # handle the nested custom task case
    # Need to be verified: nested custom task with tasks result as parameters
    nested_custom_tasks = []
    custom_task_crs_namelist = []
    for custom_task_key in custom_task.keys():
        custom_task_crs_namelist.append(custom_task_key)
    for custom_task_key in custom_task.keys():
        for inner_task_name in custom_task[custom_task_key]['task_list']:
            inner_task_name_trimmed = copy.copy(inner_task_name)
            if len(group_names[-1]) <= LOOP_GROUP_NAME_LENGTH:
                inner_task_name_trimmed = sanitize_k8s_name(inner_task_name, max_length=LOOP_GROUP_NAME_LENGTH, rev_truncate=True)
            inner_task_cr_name = '-'.join(group_names[:-1] + [inner_task_name_trimmed])
            if inner_task_cr_name in custom_task_crs_namelist:
                nested_custom_tasks.append({
                    "father_ct": custom_task_key,
                    "nested_custom_task": inner_task_cr_name
                })
    # Summary out all of the nested tasks relationships.
    for nested_custom_task in nested_custom_tasks:
        father_ct_name = nested_custom_task['father_ct']
        relationships = find_ancestors(nested_custom_tasks, father_ct_name, [], father_ct_name)
        nested_custom_task['ancestors'] = relationships['ancestors']
        nested_custom_task['root_ct'] = relationships['root_ct']

    for nested_custom_task in nested_custom_tasks:
        nested_custom_task_spec = custom_task[nested_custom_task['nested_custom_task']]['spec']
        for custom_task_cr in custom_task_crs:
            if custom_task_cr['metadata']['name'] == nested_custom_task['father_ct']:
                # handle parameters of nested custom task
                params_nested_custom_task = nested_custom_task_spec['params']
                # nested_custom_task_special_params = the global params that doesn't defined in parent custom task
                nested_custom_task_special_params = [
                    param for param in params_nested_custom_task
                    if '$(params.' in param['value'] and not bool([
                        True for ct_param in custom_task_cr['spec']['pipelineSpec']['params']
                        if param['name'] in ct_param['name']
                    ])
                ]
                custom_task_cr['spec']['pipelineSpec']['params'].extend([
                    {'name': param['name'], 'type': 'string'}for param in nested_custom_task_special_params
                ])

                if nested_custom_task['ancestors']:
                    for custom_task_cr_again in custom_task_crs:
                        if custom_task_cr_again['metadata']['name'] in nested_custom_task[
                            'ancestors'] or custom_task_cr_again['metadata']['name'] == nested_custom_task['root_ct']:
                            custom_task_cr_again['spec']['pipelineSpec']['params'].extend([
                                {'name': param['name'], 'type': 'string'}for param in nested_custom_task_special_params
                            ])
                            custom_task_cr_again['spec']['pipelineSpec']['params'] = sorted(
                                custom_task_cr_again['spec']['pipelineSpec']['params'], key=lambda k: k['name'])
                # add children params to the root tasks
                global_task_values = set()
                all_nested_loop = []
                nested_task = nested_custom_task['nested_custom_task']
                while nested_task:
                    all_nested_loop.append(nested_task)
                    has_nested_task = False
                    for n_task in nested_custom_tasks:
                        if n_task['father_ct'] == nested_task:
                            nested_task = n_task['nested_custom_task']
                            has_nested_task = True
                            break
                    if not has_nested_task:
                        break
                nested_loop_counter_params = []
                # Fetch nested loop counter params so that it won't find the nested parameters from
                # global param level.
                for task in custom_task_crs:
                    if task['metadata']['name'] in all_nested_loop:
                        # add additional pipelinerun controller generated params for the compiler to ignore for upward passing
                        iterate_param_list = ['iterateNumeric', 'iterateParam', 'iterateParamStringSeparator', 'IterationNumberParam']
                        for iterate_name in iterate_param_list:
                            if task['spec'].get(iterate_name):
                                nested_loop_counter_params.append(task['spec'].get(iterate_name))
                        
                for task in tasks:
                    if task['name'] == nested_custom_task['root_ct']:
                        task['params'].extend(copy.deepcopy(nested_custom_task_special_params))
                    elif task['name'] in nested_custom_task['ancestors'] or task[
                        'name'] == nested_custom_task['father_ct']:
                        task['params'].extend(nested_custom_task_special_params)
                    if task.get('params') is not None:
                        task['params'] = sorted(task['params'], key=lambda k: k['name'])
                        if task['name'] in all_nested_loop:
                            for param in task['params']:
                                if '$(params.' in param['value'] and 'subvar-' not in param['value']:
                                    global_task_values.add(param['value'])
                # Add any pipeline global params to the nested loop layers
                all_params = []
                for custom_param in custom_task_cr['spec']['pipelineSpec']['params']:
                    all_params.append(''.join(['$(params.', custom_param['name'], ')']))
                for global_task_value in global_task_values:
                    if global_task_value not in all_params and \
                        re.findall('\$\(params.([^ \t\n.:,;\{\}]+)\)', global_task_value)[0] not in nested_loop_counter_params:
                        all_params.append(global_task_value)
                        custom_task_cr['spec']['pipelineSpec']['params'].append(
                            {'name': re.findall('\$\(params.([^ \t\n.:,;\{\}]+)\)', global_task_value)[0],
                             'type': 'string'}
                        )
                        for task in tasks:
                            if task['name'] == nested_custom_task['father_ct']:
                                task['params'].append(
                                    {'name': re.findall('\$\(params.([^ \t\n.:,;\{\}]+)\)', global_task_value)[0],
                                     'value': global_task_value}
                                )
                for special_param in nested_custom_task_special_params:
                    for nested_param in nested_custom_task_spec['params']:
                        if nested_param['name'] == special_param['name']:
                            nested_param['value'] = '$(params.%s)' % nested_param['name']
                # need process parameters to replace results
                custom_task_cr_task_names = [cr_task['name'] for cr_task in custom_task_cr['spec']['pipelineSpec']['tasks']]
                for nested_custom_task_param in nested_custom_task_spec['params']:
                    if '$(tasks.' in nested_custom_task_param['value']:
                        param_results = re.findall('\$\(tasks.([^ \t\n.:,;\{\}]+).results.([^ \t\n.:,;\{\}]+)\)',
                                                    nested_custom_task_param['value'])
                        for param_result in param_results:
                            if param_result[0] not in custom_task_cr_task_names:
                                custom_task_cr_param_names = [p['name'] for p in custom_task_cr['spec']['pipelineSpec']['params']]
                                if nested_custom_task_param['name'] not in custom_task_cr_param_names:
                                    for index, param in enumerate(nested_custom_task_spec['params']):
                                        if nested_custom_task_param['name'] == param['name']:
                                            nested_custom_task_spec['params'].pop(index)
                                            break
                                else:
                                    nested_custom_task_spec = json.loads(
                                        json.dumps(nested_custom_task_spec).replace(nested_custom_task_param['value'],
                                        '$(params.%s)' % nested_custom_task_param['name']))
                # add nested custom task spec to main custom task
                custom_task_cr['spec']['pipelineSpec']['tasks'].append(nested_custom_task_spec)
                custom_task_cr['spec']['pipelineSpec']['params'] = sorted(
                    custom_task_cr['spec']['pipelineSpec']['params'], key=lambda k: k['name'])

    # remove the tasks belong to custom task from main workflow
    task_name_prefix = '-'.join(group_names[:-1] + [""])
    for task in tasks:
        if task['name'].replace(task_name_prefix, "") not in task_list:
            task_list_trimmed = [sanitize_k8s_name(task, max_length=LOOP_GROUP_NAME_LENGTH, rev_truncate=True) for task in task_list]
            if task['name'].replace(task_name_prefix, "") not in task_list_trimmed:
                new_tasks.append(task)
    workflow['spec']['pipelineSpec']['tasks'] = new_tasks
    # clean up any redundant parameters
    for i, task in enumerate(workflow['spec']['pipelineSpec']['tasks']):
        searched_params = []
        if task.get('params'):
            new_params_spec = []
            for workflow_param in task['params']:
                if workflow_param['name'] not in searched_params:
                    new_params_spec.append(workflow_param)
                    searched_params.append(workflow_param['name'])
            workflow['spec']['pipelineSpec']['tasks'][i]['params'] = new_params_spec
    # call post-hook api if there is
    updated_custom_task_crs = []
    for cr in custom_task_crs:
        cr_task = custom_task.get(cr['metadata']['name'])
        updated_custom_task_crs.append(
            cr_task.get('_data').post_task_spec(cr)
            if cr_task and cr_task.get('_data') and
                isinstance(cr_task.get('_data'), AddOnGroup)
            else cr)
    return updated_custom_task_crs, workflow


def find_ancestors(nested_custom_tasks: list, father_ct_name, ancestors: list, root_ct):
    """ Finds the ancestors of a nested custom task.

        Parameters
        ----------
        nested_custom_tasks : list
            A list of dictionaries, each dictionary containing the name of a nested custom task and the name of its father custom task.
        father_ct_name : str
            The name of the father custom task.
        ancestors : list
            A list of the ancestors of the nested custom task.
        root_ct : str
            The name of the root custom task.

        Returns
        -------
        relationship : dict
            A dictionary containing the ancestors of the nested custom task and the name of the root custom task.
    """
    relationship = {'ancestors': ancestors, 'root_ct': root_ct}
    for custom_task in nested_custom_tasks:
        if father_ct_name == custom_task['nested_custom_task']:
            father_ct_name = custom_task['father_ct']
            relationship = find_ancestors(nested_custom_tasks, father_ct_name, ancestors, father_ct_name)
            if relationship['root_ct'] != father_ct_name:
                ancestors.append(father_ct_name)
    return relationship



# Copyright 2021 kubeflow.org
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from kfp import dsl
from kubernetes.client.models import V1SecurityContext, V1Volume, V1PersistentVolumeClaimSpec
from typing import Dict

TEKTON_PIPELINE_ANNOTATIONS = ['sidecar.istio.io/inject', 'tekton.dev/artifact_bucket',
                               'tekton.dev/artifact_endpoint', 'tekton.dev/artifact_endpoint_scheme',
                               'tekton.dev/artifact_items', 'tekton.dev/input_artifacts', 'tekton.dev/output_artifacts']


class TektonPipelineConf(dsl.PipelineConf):
    """PipelineConf contains pipeline level settings."""

    def __init__(self, **kwargs):
        self.pipeline_labels = {}
        self.pipeline_annotations = {}
        self.tekton_inline_spec = True
        self.resource_in_separate_yaml = False
        self.security_context = None
        self.automount_service_account_token = None
        self.pipeline_env = {}
        self.pipeline_workspaces = {}
        self.generate_component_spec_annotations = True
        super().__init__(**kwargs)

    def copy(self):
        return TektonPipelineConf()\
            .set_tekton_inline_spec(self.tekton_inline_spec)\
            .set_resource_in_separate_yaml(self.resource_in_separate_yaml)\
            .set_security_context(self.security_context)\
            .set_automount_service_account_token(self.automount_service_account_token)\
            .set_pipeline_env(self.pipeline_env)

    def add_pipeline_label(self, label_name: str, value: str):
        self.pipeline_labels[label_name] = value
        return self

    def add_pipeline_annotation(self, annotation_name: str, value: str):
        if annotation_name in TEKTON_PIPELINE_ANNOTATIONS:
            raise ValueError('Cannot add pipeline annotation %s:%s because it is a reserved Tekton annotation.'
                             % annotation_name, value)
        self.pipeline_annotations[annotation_name] = value
        return self

    def set_tekton_inline_spec(self, value: bool):
        self.tekton_inline_spec = value
        return self

    def set_resource_in_separate_yaml(self, value: bool):
        self.resource_in_separate_yaml = value
        return self

    def set_security_context(self, value: V1SecurityContext):
        self.security_context = value
        return self

    def set_automount_service_account_token(self, value: bool):
        self.automount_service_account_token = value
        return self

    def add_pipeline_env(self, env_name: str, value: str):
        self.pipeline_env[env_name] = value
        return self

    def set_pipeline_env(self, value: Dict[str, str]):
        self.pipeline_env = value
        return self

    def add_pipeline_workspace(self,
                               workspace_name: str,
                               volume: V1Volume = None,
                               volume_claim_template_spec: V1PersistentVolumeClaimSpec = None,
                               path_prefix: str = None):
        if not (volume or volume_claim_template_spec) or (volume and volume_claim_template_spec):
            raise ValueError("Only either volume or volume_claim_templateneeds to be defined in add_pipeline_workspace(%s)" %
                             workspace_name)
        volumeSpec = volume if volume else volume_claim_template_spec
        self.pipeline_workspaces[workspace_name] = (volumeSpec, path_prefix)
        return self

    def set_generate_component_spec_annotations(self, value: bool):
        self.generate_component_spec_annotations = value
        return self



# Copyright 2019-2020 kubeflow.org.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from kubernetes import client as k8s_client


def env_from_secret(env_name, secret_name, secret_key):
    """ Creates a V1EnvVar object from a secret.

        Parameters
        ----------
        env_name : str
            The name of the environment variable.
        secret_name : str
            The name of the secret.
        secret_key : str
            The key of the secret.

        Returns
        -------
        V1EnvVar
            A V1EnvVar object.
    """
    return k8s_client.V1EnvVar(
        name=env_name,
        value_from=k8s_client.V1EnvVarSource(
            secret_key_ref=k8s_client.V1SecretKeySelector(
                name=secret_name,
                key=secret_key
            )
        )
    )



# Copyright 2020 kubeflow.org
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import tempfile
import time

from datetime import datetime
from typing import Mapping, Callable, Optional

import kfp

import kfp_tekton_server_api as kfp_server_api
from kfp_tekton_server_api import ApiException

from .compiler import TektonCompiler

import json
import logging

from kfp import dsl
from kfp.compiler._k8s_helper import sanitize_k8s_name

# Operators on scalar values. Only applies to one of |int_value|,
# |long_value|, |string_value| or |timestamp_value|.
_FILTER_OPERATIONS = {
    'UNKNOWN': 0,
    'EQUALS': 1,
    'NOT_EQUALS': 2,
    'GREATER_THAN': 3,
    'GREATER_THAN_EQUALS': 5,
    'LESS_THAN': 6,
    'LESS_THAN_EQUALS': 7
}

KF_PIPELINES_ENDPOINT_ENV = 'KF_PIPELINES_ENDPOINT'
KF_PIPELINES_UI_ENDPOINT_ENV = 'KF_PIPELINES_UI_ENDPOINT'
KF_PIPELINES_DEFAULT_EXPERIMENT_NAME = 'KF_PIPELINES_DEFAULT_EXPERIMENT_NAME'
KF_PIPELINES_OVERRIDE_EXPERIMENT_NAME = 'KF_PIPELINES_OVERRIDE_EXPERIMENT_NAME'
KF_PIPELINES_IAP_OAUTH2_CLIENT_ID_ENV = 'KF_PIPELINES_IAP_OAUTH2_CLIENT_ID'
KF_PIPELINES_APP_OAUTH2_CLIENT_ID_ENV = 'KF_PIPELINES_APP_OAUTH2_CLIENT_ID'
KF_PIPELINES_APP_OAUTH2_CLIENT_SECRET_ENV = 'KF_PIPELINES_APP_OAUTH2_CLIENT_SECRET'


def _add_generated_apis(target_struct, api_module, api_client):
    """Initializes a hierarchical API object based on the generated API module.
    PipelineServiceApi.create_pipeline becomes
    target_struct.pipelines.create_pipeline
    """
    Struct = type('Struct', (), {})

    def camel_case_to_snake_case(name):
        import re
        return re.sub('([a-z0-9])([A-Z])', r'\1_\2', name).lower()

    for api_name in dir(api_module):
        if not api_name.endswith('ServiceApi'):
            continue

        short_api_name = camel_case_to_snake_case(
            api_name[0:-len('ServiceApi')]) + 's'
        api_struct = Struct()
        setattr(target_struct, short_api_name, api_struct)
        service_api = getattr(api_module.api, api_name)
        initialized_service_api = service_api(api_client)
        for member_name in dir(initialized_service_api):
            if member_name.startswith('_') or member_name.endswith(
                    '_with_http_info'):
                continue

            bound_member = getattr(initialized_service_api, member_name)
            setattr(api_struct, member_name, bound_member)
    models_struct = Struct()
    for member_name in dir(api_module.models):
        if not member_name[0].islower():
            setattr(models_struct, member_name,
                    getattr(api_module.models, member_name))
    target_struct.api_models = models_struct


class TektonClient(kfp.Client):
    """Tekton API Client for Kubeflow Pipelines."""

    def __init__(self,
                 host=None,
                 client_id=None,
                 namespace='kubeflow',
                 other_client_id=None,
                 other_client_secret=None,
                 existing_token=None,
                 cookies=None,
                 proxy=None,
                 ssl_ca_cert=None,
                 kube_context=None,
                 credentials=None,
                 ui_host=None):
        """Create a new instance of kfp client."""
        host = host or os.environ.get(KF_PIPELINES_ENDPOINT_ENV)
        self._uihost = os.environ.get(KF_PIPELINES_UI_ENDPOINT_ENV, ui_host or
                                      host)
        client_id = client_id or os.environ.get(
            KF_PIPELINES_IAP_OAUTH2_CLIENT_ID_ENV)
        other_client_id = other_client_id or os.environ.get(
            KF_PIPELINES_APP_OAUTH2_CLIENT_ID_ENV)
        other_client_secret = other_client_secret or os.environ.get(
            KF_PIPELINES_APP_OAUTH2_CLIENT_SECRET_ENV)

        config = self._load_config(host, client_id, namespace, other_client_id,
                                   other_client_secret, existing_token, proxy,
                                   ssl_ca_cert, kube_context, credentials)
        # Save the loaded API client configuration, as a reference if update is
        # needed.
        self._load_context_setting_or_default()

        # If custom namespace provided, overwrite the loaded or default one in
        # context settings for current client instance
        if namespace != 'kubeflow':
            self._context_setting['namespace'] = namespace

        self._existing_config = config
        if cookies is None:
            cookies = self._context_setting.get('client_authentication_cookie')
        api_client = kfp_server_api.api_client.ApiClient(
            config,
            cookie=cookies,
            header_name=self._context_setting.get(
                'client_authentication_header_name'),
            header_value=self._context_setting.get(
                'client_authentication_header_value'))
        _add_generated_apis(self, kfp_server_api, api_client)
        self._job_api = kfp_server_api.api.job_service_api.JobServiceApi(
            api_client)
        self._run_api = kfp_server_api.api.run_service_api.RunServiceApi(
            api_client)
        self._experiment_api = kfp_server_api.api.experiment_service_api.ExperimentServiceApi(
            api_client)
        self._pipelines_api = kfp_server_api.api.pipeline_service_api.PipelineServiceApi(
            api_client)
        self._upload_api = kfp_server_api.api.PipelineUploadServiceApi(
            api_client)
        self._healthz_api = kfp_server_api.api.healthz_service_api.HealthzServiceApi(
            api_client)
        if not self._context_setting['namespace'] and self.get_kfp_healthz(
        ).multi_user is True:
            try:
                with open(kfp.Client.NAMESPACE_PATH, 'r') as f:
                    current_namespace = f.read()
                    self.set_user_namespace(current_namespace)
            except FileNotFoundError:
                logging.info(
                    'Failed to automatically set namespace.', exc_info=False)
                
    def wait_for_run_completion(self, run_id: str, timeout: int):
        """Waits for a run to complete.

        Args:
          run_id: Run id, returned from run_pipeline.
          timeout: Timeout in seconds.

        Returns:
          A run detail object: Most important fields are run and pipeline_runtime.

        Raises:
          TimeoutError: if the pipeline run failed to finish before the specified timeout.
        """
        status = 'Running:'
        start_time = datetime.datetime.now()
        if isinstance(timeout, datetime.timedelta):
            timeout = timeout.total_seconds()
        is_valid_token = False
        # Tekton pipelineruns Status
        # List of Tekton status: https://github.com/tektoncd/pipeline/blob/main/pkg/apis/pipeline/v1/pipelinerun_types.go
        # List of Tekton error status: https://github.com/tektoncd/pipeline/blob/main/pkg/reconciler/pipelinerun/pipelinerun.go
        tekton_completion_status = {'succeeded',
                                    'failed',
                                    'skipped',
                                    'error',
                                    'completed',
                                    'pipelineruncancelled',
                                    'pipelineruncouldntcancel',
                                    'pipelineruntimeout',
                                    'cancelled',
                                    'stoppedrunfinally',
                                    'cancelledrunfinally',
                                    'invalidtaskresultreference'}
        while True:
            try:
                get_run_response = self._run_api.get_run(run_id=run_id)
                is_valid_token = True
            except ApiException as api_ex:
                # if the token is valid but receiving 401 Unauthorized error
                # then refresh the token
                if is_valid_token and api_ex.status == 401:
                    logging.info('Access token has expired !!! Refreshing ...')
                    self._refresh_api_client_token()
                    continue
                else:
                    raise api_ex
            status = get_run_response.run.status
            elapsed_time = (datetime.datetime.now() -
                            start_time).total_seconds()
            logging.info('Waiting for the job to complete...')
            if elapsed_time > timeout:
                raise TimeoutError('Run timeout')
            if status is not None and status.lower() in tekton_completion_status:
                return get_run_response
            time.sleep(5)

    def create_experiment(
            self,
            name: str,
            description: str = None,
            namespace: str = None) -> kfp_server_api.V1Experiment:
        """Create a new experiment.

        Args:
          name: The name of the experiment.
          description: Description of the experiment.
          namespace: Kubernetes namespace where the experiment should be created.
            For single user deployment, leave it as None;
            For multi user, input a namespace where the user is authorized.

        Returns:
          An Experiment object. Most important field is id.
        """
        namespace = namespace or self.get_user_namespace()
        experiment = None
        try:
            experiment = self.get_experiment(
                experiment_name=name, namespace=namespace)
        except ValueError as error:
            # Ignore error if the experiment does not exist.
            if not str(error).startswith('No experiment is found with name'):
                raise error

        if not experiment:
            logging.info('Creating experiment {}.'.format(name))

            resource_references = []
            if namespace:
                key = kfp_server_api.models.V1ResourceKey(
                    id=namespace,
                    type=kfp_server_api.models.V1ResourceType.NAMESPACE)
                reference = kfp_server_api.models.V1ResourceReference(
                    key=key,
                    relationship=kfp_server_api.models.V1Relationship.OWNER)
                resource_references.append(reference)

            experiment = kfp_server_api.models.V1Experiment(
                name=name,
                description=description,
                resource_references=resource_references)
            experiment = self._experiment_api.create_experiment(body=experiment)

        if self._is_ipython():
            import IPython
            html = \
                ('<a href="%s/#/experiments/details/%s" target="_blank" >Experiment details</a>.'
                % (self._get_url_prefix(), experiment.id))
            IPython.display.display(IPython.display.HTML(html))
        return experiment

    def list_experiments(
            self,
            page_token='',
            page_size=10,
            sort_by='',
            namespace=None,
            filter=None) -> kfp_server_api.V1ListExperimentsResponse:
        """List experiments.

        Args:
          page_token: Token for starting of the page.
          page_size: Size of the page.
          sort_by: Can be '[field_name]', '[field_name] desc'. For example, 'name desc'.
          namespace: Kubernetes namespace where the experiment was created.
            For single user deployment, leave it as None;
            For multi user, input a namespace where the user is authorized.
          filter: A url-encoded, JSON-serialized Filter protocol buffer
            (see [filter.proto](https://github.com/kubeflow/pipelines/blob/master/backend/api/filter.proto)).

        Returns:
          A response object including a list of experiments and next page token.
        """
        namespace = namespace or self.get_user_namespace()
        response = self._experiment_api.list_experiment(
            page_token=page_token,
            page_size=page_size,
            sort_by=sort_by,
            resource_reference_key_type=kfp_server_api.models.v1_resource_type
            .V1ResourceType.NAMESPACE,
            resource_reference_key_id=namespace,
            filter=filter)
        return response

    def get_experiment(self,
                       experiment_id=None,
                       experiment_name=None,
                       namespace=None) -> kfp_server_api.V1Experiment:
        """Get details of an experiment.

        Either experiment_id or experiment_name is required

        Args:
          experiment_id: Id of the experiment. (Optional)
          experiment_name: Name of the experiment. (Optional)
          namespace: Kubernetes namespace where the experiment was created.
            For single user deployment, leave it as None;
            For multi user, input the namespace where the user is authorized.

        Returns:
          A response object including details of a experiment.

        Raises:
          kfp_server_api.ApiException: If experiment is not found or None of the arguments is provided
        """
        namespace = namespace or self.get_user_namespace()
        if experiment_id is None and experiment_name is None:
            raise ValueError(
                'Either experiment_id or experiment_name is required')
        if experiment_id is not None:
            return self._experiment_api.get_experiment(id=experiment_id)
        experiment_filter = json.dumps({
            'predicates': [{
                'op': _FILTER_OPERATIONS['EQUALS'],
                'key': 'name',
                'stringValue': experiment_name,
            }]
        })
        if namespace:
            result = self._experiment_api.list_experiment(
                filter=experiment_filter,
                resource_reference_key_type=kfp_server_api.models
                .v1_resource_type.V1ResourceType.NAMESPACE,
                resource_reference_key_id=namespace)
        else:
            result = self._experiment_api.list_experiment(
                filter=experiment_filter)
        if not result.experiments:
            raise ValueError(
                'No experiment is found with name {}.'.format(experiment_name))
        if len(result.experiments) > 1:
            raise ValueError(
                'Multiple experiments is found with name {}.'.format(
                    experiment_name))
        return result.experiments[0]

    # TODO: provide default namespace, similar to kubectl default namespaces.
    def run_pipeline(
        self,
        experiment_id: str,
        job_name: str,
        pipeline_package_path: Optional[str] = None,
        params: Optional[dict] = None,
        pipeline_id: Optional[str] = None,
        version_id: Optional[str] = None,
        pipeline_root: Optional[str] = None,
        enable_caching: Optional[str] = None,
        service_account: Optional[str] = None,
    ) -> kfp_server_api.V1Run:
        """Run a specified pipeline.

        Args:
          experiment_id: The id of an experiment.
          job_name: Name of the job.
          pipeline_package_path: Local path of the pipeline package
                                 (the filename should end with one of the following .tar.gz, .tgz, .zip, .yaml, .yml).
          params: A dictionary with key (string) as param name and value (string) as as param value.
          pipeline_id: The id of a pipeline.
          version_id: The id of a pipeline version.
            If both pipeline_id and version_id are specified, version_id will take precendence.
            If only pipeline_id is specified, the default version of this pipeline is used to create the run.
          pipeline_root: The root path of the pipeline outputs. This argument should
            be used only for pipeline compiled with
            dsl.PipelineExecutionMode.V2_COMPATIBLE or
            dsl.PipelineExecutionMode.V2_ENGINGE mode.
          enable_caching: Optional. Whether or not to enable caching for the run.
            This setting affects v2 compatible mode and v2 mode only.
            If not set, defaults to the compile time settings, which are True for all
            tasks by default, while users may specify different caching options for
            individual tasks.
            If set, the setting applies to all tasks in the pipeline -- overrides
            the compile time settings.
          service_account: Optional. Specifies which Kubernetes service account this
            run uses.

        Returns:
          A run object. Most important field is id.
        """
        if params is None:
            params = {}

        if pipeline_root is not None:
            params[dsl.ROOT_PARAMETER_NAME] = pipeline_root

        job_config = self._create_job_config(
            experiment_id=experiment_id,
            params=params,
            pipeline_package_path=pipeline_package_path,
            pipeline_id=pipeline_id,
            version_id=version_id,
            enable_caching=enable_caching,
        )
        run_body = kfp_server_api.models.V1Run(
            pipeline_spec=job_config.spec,
            resource_references=job_config.resource_references,
            name=job_name,
            service_account=service_account)

        response = self._run_api.create_run(body=run_body)

        if self._is_ipython():
            import IPython
            html = (
                '<a href="%s/#/runs/details/%s" target="_blank" >Run details</a>.'
                % (self._get_url_prefix(), response.run.id))
            IPython.display.display(IPython.display.HTML(html))
        return response.run

    def create_recurring_run(
        self,
        experiment_id: str,
        job_name: str,
        description: Optional[str] = None,
        start_time: Optional[str] = None,
        end_time: Optional[str] = None,
        interval_second: Optional[int] = None,
        cron_expression: Optional[str] = None,
        max_concurrency: Optional[int] = 1,
        no_catchup: Optional[bool] = None,
        params: Optional[dict] = None,
        pipeline_package_path: Optional[str] = None,
        pipeline_id: Optional[str] = None,
        version_id: Optional[str] = None,
        enabled: bool = True,
        enable_caching: Optional[bool] = None,
        service_account: Optional[str] = None,
    ) -> kfp_server_api.V1Job:
        """Create a recurring run.

        Args:
          experiment_id: The string id of an experiment.
          job_name: Name of the job.
          description: An optional job description.
          start_time: The RFC3339 time string of the time when to start the job.
          end_time: The RFC3339 time string of the time when to end the job.
          interval_second: Integer indicating the seconds between two recurring runs in for a periodic schedule.
          cron_expression: A cron expression representing a set of times, using 6 space-separated fields, e.g. "0 0 9 ? * 2-6".
            See `here <https://pkg.go.dev/github.com/robfig/cron#hdr-CRON_Expression_Format>`_ for details of the cron expression format.
          max_concurrency: Integer indicating how many jobs can be run in parallel.
          no_catchup: Whether the recurring run should catch up if behind schedule.
            For example, if the recurring run is paused for a while and re-enabled
            afterwards. If no_catchup=False, the scheduler will catch up on (backfill) each
            missed interval. Otherwise, it only schedules the latest interval if more than one interval
            is ready to be scheduled.
            Usually, if your pipeline handles backfill internally, you should turn catchup
            off to avoid duplicate backfill. (default: {False})
          pipeline_package_path: Local path of the pipeline package
                                 (the filename should end with one of the following .tar.gz, .tgz, .zip, .yaml, .yml).
          params: A dictionary with key (string) as param name and value (string) as param value.
          pipeline_id: The id of a pipeline.
          version_id: The id of a pipeline version.
            If both pipeline_id and version_id are specified, version_id will take precendence.
            If only pipeline_id is specified, the default version of this pipeline is used to create the run.
          enabled: A bool indicating whether the recurring run is enabled or disabled.
          enable_caching: Optional. Whether or not to enable caching for the run.
            This setting affects v2 compatible mode and v2 mode only.
            If not set, defaults to the compile time settings, which are True for all
            tasks by default, while users may specify different caching options for
            individual tasks.
            If set, the setting applies to all tasks in the pipeline -- overrides
            the compile time settings.
          service_account: Optional. Specifies which Kubernetes service account this
            recurring run uses.

        Returns:
          A Job object. Most important field is id.

        Raises:
          ValueError: If required parameters are not supplied.
        """

        job_config = self._create_job_config(
            experiment_id=experiment_id,
            params=params,
            pipeline_package_path=pipeline_package_path,
            pipeline_id=pipeline_id,
            version_id=version_id,
            enable_caching=enable_caching,
        )

        if all([interval_second, cron_expression
               ]) or not any([interval_second, cron_expression]):
            raise ValueError(
                'Either interval_second or cron_expression is required')
        if interval_second is not None:
            trigger = kfp_server_api.models.V1Trigger(
                periodic_schedule=kfp_server_api.models.V1PeriodicSchedule(
                    start_time=start_time,
                    end_time=end_time,
                    interval_second=interval_second))
        if cron_expression is not None:
            trigger = kfp_server_api.models.V1Trigger(
                cron_schedule=kfp_server_api.models.V1CronSchedule(
                    start_time=start_time,
                    end_time=end_time,
                    cron=cron_expression))

        job_body = kfp_server_api.models.V1Job(
            enabled=enabled,
            pipeline_spec=job_config.spec,
            resource_references=job_config.resource_references,
            name=job_name,
            description=description,
            no_catchup=no_catchup,
            trigger=trigger,
            max_concurrency=max_concurrency,
            service_account=service_account)
        return self._job_api.create_job(body=job_body)

    def _create_job_config(
        self,
        experiment_id: str,
        params: Optional[dict],
        pipeline_package_path: Optional[str],
        pipeline_id: Optional[str],
        version_id: Optional[str],
        enable_caching: Optional[bool],
    ):
        """Create a JobConfig with spec and resource_references.

        Args:
          experiment_id: The id of an experiment.
          pipeline_package_path: Local path of the pipeline package
                                 (the filename should end with one of the following .tar.gz, .tgz, .zip, .yaml, .yml).
          params: A dictionary with key (string) as param name and value (string) as param value.
          pipeline_id: The id of a pipeline.
          version_id: The id of a pipeline version.
            If both pipeline_id and version_id are specified, version_id will take precendence.
            If only pipeline_id is specified, the default version of this pipeline is used to create the run.
          enable_caching: Whether or not to enable caching for the run.
            This setting affects v2 compatible mode and v2 mode only.
            If not set, defaults to the compile time settings, which are True for all
            tasks by default, while users may specify different caching options for
            individual tasks.
            If set, the setting applies to all tasks in the pipeline -- overrides
            the compile time settings.

        Returns:
          A JobConfig object with attributes spec and resource_reference.
        """

        class JobConfig:

            def __init__(self, spec, resource_references):
                self.spec = spec
                self.resource_references = resource_references

        params = params or {}
        pipeline_json_string = None
        if pipeline_package_path:
            pipeline_obj = self._extract_pipeline_yaml(pipeline_package_path)

            # Caching option set at submission time overrides the compile time settings.
            if enable_caching is not None:
                self._override_caching_options(pipeline_obj, enable_caching)

            pipeline_json_string = json.dumps(pipeline_obj)
        api_params = [
            kfp_server_api.V1Parameter(
                name=sanitize_k8s_name(name=k, allow_capital_underscore=True),
                value=str(v) if type(v) not in (list, dict) else json.dumps(v))
            for k, v in params.items()
        ]
        resource_references = []
        key = kfp_server_api.models.V1ResourceKey(
            id=experiment_id,
            type=kfp_server_api.models.V1ResourceType.EXPERIMENT)
        reference = kfp_server_api.models.V1ResourceReference(
            key=key, relationship=kfp_server_api.models.V1Relationship.OWNER)
        resource_references.append(reference)

        if version_id:
            key = kfp_server_api.models.V1ResourceKey(
                id=version_id,
                type=kfp_server_api.models.V1ResourceType.PIPELINE_VERSION)
            reference = kfp_server_api.models.V1ResourceReference(
                key=key,
                relationship=kfp_server_api.models.V1Relationship.CREATOR)
            resource_references.append(reference)

        spec = kfp_server_api.models.V1PipelineSpec(
            pipeline_id=pipeline_id,
            workflow_manifest=pipeline_json_string,
            parameters=api_params)
        return JobConfig(spec=spec, resource_references=resource_references)

    def list_runs(self,
                  page_token='',
                  page_size=10,
                  sort_by='',
                  experiment_id=None,
                  namespace=None,
                  filter=None) -> kfp_server_api.V1ListRunsResponse:
        """List runs, optionally can be filtered by experiment or namespace.

        Args:
          page_token: Token for starting of the page.
          page_size: Size of the page.
          sort_by: One of 'field_name', 'field_name desc'. For example, 'name desc'.
          experiment_id: Experiment id to filter upon
          namespace: Kubernetes namespace to filter upon.
            For single user deployment, leave it as None;
            For multi user, input a namespace where the user is authorized.
          filter: A url-encoded, JSON-serialized Filter protocol buffer
            (see [filter.proto](https://github.com/kubeflow/pipelines/blob/master/backend/api/filter.proto)).

        Returns:
          A response object including a list of experiments and next page token.
        """
        namespace = namespace or self.get_user_namespace()
        if experiment_id is not None:
            response = self._run_api.list_runs(
                page_token=page_token,
                page_size=page_size,
                sort_by=sort_by,
                resource_reference_key_type=kfp_server_api.models
                .v1_resource_type.V1ResourceType.EXPERIMENT,
                resource_reference_key_id=experiment_id,
                filter=filter)
        elif namespace:
            response = self._run_api.list_runs(
                page_token=page_token,
                page_size=page_size,
                sort_by=sort_by,
                resource_reference_key_type=kfp_server_api.models
                .v1_resource_type.V1ResourceType.NAMESPACE,
                resource_reference_key_id=namespace,
                filter=filter)
        else:
            response = self._run_api.list_runs(
                page_token=page_token,
                page_size=page_size,
                sort_by=sort_by,
                filter=filter)
        return response

    def list_recurring_runs(self,
                            page_token='',
                            page_size=10,
                            sort_by='',
                            experiment_id=None,
                            filter=None) -> kfp_server_api.V1ListJobsResponse:
        """List recurring runs.

        Args:
          page_token: Token for starting of the page.
          page_size: Size of the page.
          sort_by: One of 'field_name', 'field_name desc'. For example, 'name desc'.
          experiment_id: Experiment id to filter upon.
          filter: A url-encoded, JSON-serialized Filter protocol buffer
            (see [filter.proto](https://github.com/kubeflow/pipelines/blob/master/backend/api/filter.proto)).

        Returns:
          A response object including a list of recurring_runs and next page token.
        """
        if experiment_id is not None:
            response = self._job_api.list_jobs(
                page_token=page_token,
                page_size=page_size,
                sort_by=sort_by,
                resource_reference_key_type=kfp_server_api.models
                .v1_resource_type.V1ResourceType.EXPERIMENT,
                resource_reference_key_id=experiment_id,
                filter=filter)
        else:
            response = self._job_api.list_jobs(
                page_token=page_token,
                page_size=page_size,
                sort_by=sort_by,
                filter=filter)
        return response

    def list_pipeline_versions(
            self,
            pipeline_id: str,
            page_token: str = '',
            page_size: int = 10,
            sort_by: str = ''
    ) -> kfp_server_api.V1ListPipelineVersionsResponse:
        """Lists pipeline versions.

        Args:
          pipeline_id: Id of the pipeline to list versions
          page_token: Token for starting of the page.
          page_size: Size of the page.
          sort_by: One of 'field_name', 'field_name desc'. For example, 'name desc'.

        Returns:
          A response object including a list of versions and next page token.

        Raises:
          kfp_server_api.ApiException: If pipeline is not found.
        """

        return self._pipelines_api.list_pipeline_versions(
            page_token=page_token,
            page_size=page_size,
            sort_by=sort_by,
            resource_key_type=kfp_server_api.models.v1_resource_type
            .V1ResourceType.PIPELINE,
            resource_key_id=pipeline_id)

    def create_run_from_pipeline_func(self,
                                      pipeline_func: Callable,
                                      arguments: Mapping[str, str],
                                      run_name=None,
                                      experiment_name=None,
                                      pipeline_conf: kfp.dsl.PipelineConf = None,
                                      namespace=None):
      """Runs pipeline on Kubernetes cluster with Kubeflow Pipelines Tekton backend.

      This command compiles the pipeline function, creates or gets an experiment and
      submits the pipeline for execution.

      :param pipeline_func: A function that describes a pipeline by calling components
          and composing them into execution graph.
      :param arguments: Arguments to the pipeline function provided as a dict.
      :param run_name: Optional. Name of the run to be shown in the UI.
      :param experiment_name: Optional. Name of the experiment to add the run to.
      :param pipeline_conf: Optional. Pipeline configuration.
      :param namespace: kubernetes namespace where the pipeline runs are created.
          For single user deployment, leave it as None;
          For multi user, input a namespace where the user is authorized
      :return: RunPipelineResult
      """

      # TODO: Check arguments against the pipeline function
      pipeline_name = pipeline_func.__name__
      run_name = run_name or pipeline_name + ' ' + datetime.now().strftime('%Y-%m-%d %H-%M-%S')
      try:
          (_, pipeline_package_path) = tempfile.mkstemp(suffix='.zip')
          TektonCompiler().compile(pipeline_func, pipeline_package_path, pipeline_conf=pipeline_conf)
          return self.create_run_from_pipeline_package(pipeline_package_path, arguments,
                                                      run_name, experiment_name, namespace)
      finally:
          os.remove(pipeline_package_path)



# Copyright 2020 kubeflow.org
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
import sys

import kfp.compiler.main as kfp_compiler_main
from kfp_tekton.compiler.pipeline_utils import TektonPipelineConf

from . import TektonCompiler


def parse_arguments():
  """Parse command line arguments."""

  parser = argparse.ArgumentParser()
  parser.add_argument('--py',
                      type=str,
                      help='local absolute path to a py file.')
  parser.add_argument('--package',
                      type=str,
                      help='local path to a pip installable python package file.')
  parser.add_argument('--function',
                      type=str,
                      help='The name of the function to compile if there are multiple.')
  parser.add_argument('--namespace',
                      type=str,
                      help='The namespace for the pipeline function')
  parser.add_argument('--output',
                      type=str,
                      required=True,
                      help='local path to the output workflow yaml file.')
  parser.add_argument('--disable-type-check',
                      action='store_true',
                      help='disable the type check, default is enabled.')
  parser.add_argument('--disable-task-inline',
                      action='store_true',
                      help='disable taskSpec inlining, default is enabled.')

  args = parser.parse_args()
  return args


def _compile_pipeline_function(pipeline_funcs, function_name, output_path, type_check, tekton_pipeline_conf=None):
  if len(pipeline_funcs) == 0:
    raise ValueError('A function with @dsl.pipeline decorator is required in the py file.')

  if len(pipeline_funcs) > 1 and not function_name:
    func_names = [x.__name__ for x in pipeline_funcs]
    raise ValueError('There are multiple pipelines: %s. Please specify --function.' % func_names)

  if function_name:
    pipeline_func = next((x for x in pipeline_funcs if x.__name__ == function_name), None)
    if not pipeline_func:
      raise ValueError('The function "%s" does not exist. '
                       'Did you forget @dsl.pipeline decoration?' % function_name)
  else:
    pipeline_func = pipeline_funcs[0]

  TektonCompiler().compile(pipeline_func, output_path, type_check, tekton_pipeline_conf=tekton_pipeline_conf)


def compile_pyfile(pyfile, function_name, output_path, type_check, tekton_pipeline_conf=None):
  sys.path.insert(0, os.path.dirname(pyfile))
  try:
    filename = os.path.basename(pyfile)
    with kfp_compiler_main.PipelineCollectorContext() as pipeline_funcs:
      __import__(os.path.splitext(filename)[0])
    _compile_pipeline_function(pipeline_funcs, function_name, output_path, type_check, tekton_pipeline_conf)
  finally:
    del sys.path[0]
    sys.modules.pop(os.path.splitext(filename)[0])


def main():
    args = parse_arguments()
    if ((args.py is None and args.package is None) or
        (args.py is not None and args.package is not None)):
        raise ValueError('Either --py or --package is needed but not both.')
    if args.py:
        tekton_pipeline_conf = TektonPipelineConf()
        tekton_pipeline_conf.set_tekton_inline_spec(True)
        if args.disable_task_inline:
            tekton_pipeline_conf.set_tekton_inline_spec(False)
        compile_pyfile(args.py, args.function, args.output, not args.disable_type_check, tekton_pipeline_conf)
    else:
        if args.namespace is None:
            raise ValueError('--namespace is required for compiling packages.')
        # TODO: deprecated, but without the monkey-patch this will produce Argo YAML
        kfp_compiler_main.compile_package(args.package, args.namespace, args.function, args.output, not args.disable_type_check)
